# -*- coding: utf-8 -*-
"""Anomaly detection models."""

from copy import deepcopy

# Base models and utilities
from .base import TimeSeriesParametricModel
from .forecasters import PeriodicAverageForecaster

# Setup logging
import logging
logger = logging.getLogger(__name__)

# Suppress TensorFlow warnings as default behavior
import tensorflow as tf
tf.compat.v1.logging.set_verbosity(tf.compat.v1.logging.ERROR)


#=============================
#  Generic Anomaly Detector
#=============================

class AnomalyDetector(TimeSeriesParametricModel):
    
    def _predict(self, *args, **kwargs):
        raise NotImplementedError('Anomaly detectors can be used only from the apply() method.') from None



#=============================
# Forecaster Anomaly detector
#=============================

class ForecasterAnomalyDetector(AnomalyDetector):

    @property
    def forecaster_class(self):
        raise NotImplementedError('No forecaster set for this model. Please review your code.')


    def __init__(self, path=None, id=None, **kwargs):
        
        super(ForecasterAnomalyDetector, self).__init__(path=path, id=id)

        # Load the forecaster as nested model if we have loaded the model
        if self.fitted:
            # Note: the forecaster_id is the nested folder where the forecaster is saved
            forecaster_dir = path+'/'+self.data['forecaster_id']
            self.forecaster = self.forecaster_class(forecaster_dir)
        else:
            # Initialize the forecaster              
            self.forecaster = self.forecaster_class(**kwargs)
            
        # Finally, set the id of the forecaster in the data
        self.data['forecaster_id'] = self.forecaster.data['id']
            
    def save(self, path):

        # Save the anomaly detection model
        model_dir = super(ForecasterAnomalyDetector, self).save(path)

        # ..and save the forecaster as nested model
        self.forecaster.save(model_dir)

        # Return the model dir
        return model_dir


    def __get_actual_and_predicted(self, timeseries, i, key, forecaster_window):

        # Call model predict logic and compare with the actual data
        actual    = timeseries[i].data[key]
        
        try:
            # Try the optimized predict call (which just use the data as-is)
            prediction = self.forecaster.predict(timeseries, n=1, forecast_start = i-1)
            
        except TypeError as e:
            if '_predict() got an unexpected keyword argument \'forecast_start\'' in str(e):
                # Otherwise, create on the fly a slice of the time series for the window.
                # Datapoints are only linked, not copied - so this is just a minor overhead.
                window_timeseries = timeseries[i-forecaster_window:i]
                prediction = self.forecaster.predict(window_timeseries, n=1)
            else:
                raise e

        # TODO: this is because of forecasters not supporting multi-step forecasts.
        if not isinstance(prediction, list):
            predicted = prediction[key]
        else:
            predicted = prediction[0][key]
        
        return (actual, predicted)

    def _fit(self, timeseries, *args, stdevs=3, **kwargs):

        if len(timeseries.data_keys()) > 1:
            raise NotImplementedError('Multivariate time series are not yet supported')

        # Fit the forecaster
        self.forecaster.fit(timeseries, *args, **kwargs)
        
        # Evaluate the forecaster for one step ahead and get AEs
        AEs = []
        for key in timeseries.data_keys():
            
            for i, _ in enumerate(timeseries):
                
                forecaster_window = self.forecaster.data['window']
                
                if i <=  forecaster_window:    
                    continue
                
                actual, predicted = self.__get_actual_and_predicted(timeseries, i, key, forecaster_window)
                
                AEs.append(abs(actual-predicted))

        # Compute distribution for the AEs ans set the threshold
        from scipy.stats import norm
        mean, stdev = norm.fit(AEs)
        logger.info('Using {} standard deviations as anomaly threshold: {}'.format(stdevs, stdev*stdevs))
        
        # Set AE-based threshold
        self.data['stdev'] = stdev
        self.data['stdevs'] = stdevs
        self.data['AE_threshold'] = stdev*stdevs


    def _apply(self, timeseries, inplace=False, details=False, logs=False, stdevs=None):
        
        if inplace:
            raise Exception('Anomaly detection cannot be run inplace')

        if len(timeseries.data_keys()) > 1:
            raise NotImplementedError('Multivariate time series are not yet supported')
        
        result_timeseries = timeseries.__class__()

        for key in timeseries.data_keys():
            
            for i, item in enumerate(timeseries):
                forecaster_window = self.forecaster.data['window']
                if i <=  forecaster_window:    
                    continue
                
                actual, predicted = self.__get_actual_and_predicted(timeseries, i, key, forecaster_window)
                #if logs:
                #    logger.info('{}: {} vs {}'.format(timeseries[i].dt, actual, predicted))

                AE = abs(actual-predicted)
                
                item = deepcopy(item)
                
                if stdevs:
                    AE_threshold =  self.data['stdev'] * stdevs 
                else:
                    AE_threshold =  self.data['stdev'] * self.data['stdevs'] 
                    
                if AE > AE_threshold: 
                    if logs:
                        logger.info('Detected anomaly for item starting @ {} ({}) with AE="{:.3f}..."'.format(item.t, item.dt, AE))
                    item.anomaly = 1
                else:
                    item.anomaly = 0
                
                # Add details?
                if details:
                    if isinstance(details, list):
                        if 'AE' in details:
                            item.data['AE_{}'.format(key)] = AE
                        if 'predicted' in details:
                            item.data['{}_predicted'.format(key)] = predicted
                    else:
                        item.data['AE_{}'.format(key)] = AE
                        item.data['{}_predicted'.format(key)] = predicted

                result_timeseries.append(item)
        
        return result_timeseries 



#=============================
# P. Average Anomaly detector
#=============================

class PeriodicAverageAnomalyDetector(ForecasterAnomalyDetector):

    forecaster_class = PeriodicAverageForecaster





import unittest
import os
from ..datastructures import TimePoint, DataTimePoint, DataTimeSlotSeries, DataTimeSlot
from ..datastructures import TimePointSeries, DataTimePointSeries
from ..operations import derivative, integral, diff, csum, min, max, avg, filter, select, mavg
from ..operations import Operation

# Setup logging
import logging
logging.basicConfig(level=os.environ.get('LOGLEVEL') if os.environ.get('LOGLEVEL') else 'CRITICAL')

class TestBaseOpertions(unittest.TestCase):
    
    def test_base(self):
        
        operation_from_callable = Operation()
        self.assertEqual(operation_from_callable.__name__, 'operation')
                
        def operation_as_function(data):
            pass
        
        self.assertEqual(operation_as_function.__name__, 'operation_as_function')
        

class TestMathOperations(unittest.TestCase):
  
    def test_diff_csum(self):
  
        # Test data        
        data_time_slot_series = DataTimeSlotSeries()
        data_time_slot_series.append(DataTimeSlot(start=TimePoint(0), end=TimePoint(60), data={'value':10}))
        data_time_slot_series.append(DataTimeSlot(start=TimePoint(60), end=TimePoint(120), data={'value':12}))
        data_time_slot_series.append(DataTimeSlot(start=TimePoint(120), end=TimePoint(180), data={'value':15}))
        data_time_slot_series.append(DataTimeSlot(start=TimePoint(180), end=TimePoint(240), data={'value':16}))
           
        # Test standard, from the series
        diff_data_time_slot_series = diff(data_time_slot_series)
        self.assertEqual(len(diff_data_time_slot_series), 4)
        self.assertEqual(diff_data_time_slot_series[0].data['value_diff'],0) # 2?
        self.assertEqual(diff_data_time_slot_series[1].data['value_diff'],2)
        self.assertEqual(diff_data_time_slot_series[2].data['value_diff'],3)
        self.assertEqual(diff_data_time_slot_series[3].data['value_diff'],1)

        # Test csum as well  
        diff_csum_data_time_slot_series = csum(diff_data_time_slot_series, offset=10)
        self.assertEqual(diff_csum_data_time_slot_series[0].data['value_diff_csum'], 10)
        self.assertEqual(diff_csum_data_time_slot_series[1].data['value_diff_csum'], 12)
        self.assertEqual(diff_csum_data_time_slot_series[2].data['value_diff_csum'], 15)
        self.assertEqual(diff_csum_data_time_slot_series[3].data['value_diff_csum'], 16)
        
        # Test standalone
        self.assertEqual(len(diff(data_time_slot_series)), 4)
  
        # Test in-place  behavior
        diff(data_time_slot_series, inplace=True)
        self.assertEqual(len(data_time_slot_series), 4)
        self.assertEqual(data_time_slot_series[0].data['value'],10)
        self.assertEqual(data_time_slot_series[0].data['value_diff'],0)
        self.assertEqual(data_time_slot_series[1].data['value'],12)
        self.assertEqual(data_time_slot_series[1].data['value_diff'],2)
        self.assertEqual(data_time_slot_series[2].data['value'],15)
        self.assertEqual(data_time_slot_series[2].data['value_diff'],3)
        self.assertEqual(data_time_slot_series[3].data['value'],16)
        self.assertEqual(data_time_slot_series[3].data['value_diff'],1)

        # Multi-key Test data
        data_time_point_series = DataTimePointSeries()
        data_time_point_series.append(DataTimePoint(t=0, data={'value':10, 'another_value': 75}))
        data_time_point_series.append(DataTimePoint(t=60, data={'value':12, 'another_value': 65}))
        data_time_point_series.append(DataTimePoint(t=120, data={'value':6, 'another_value': 32}))
        data_time_point_series.append(DataTimePoint(t=180, data={'value':16, 'another_value': 10}))
        diff_data_time_point_series = data_time_point_series.diff()
        self.assertEqual(diff_data_time_point_series[0].data['value_diff'],0)
        self.assertEqual(diff_data_time_point_series[0].data['another_value_diff'],0)
        self.assertEqual(diff_data_time_point_series[1].data['value_diff'],2)
        self.assertEqual(diff_data_time_point_series[1].data['another_value_diff'],-10)

        # Test offset
        diff_csum_data_time_point_series = diff_data_time_point_series.csum(offset={'value_diff':10, 'another_value_diff':75})
        self.assertEqual(diff_csum_data_time_point_series[0].data['value_diff_csum'],10)
        self.assertEqual(diff_csum_data_time_point_series[0].data['another_value_diff_csum'],75)
        self.assertEqual(diff_csum_data_time_point_series[3].data['value_diff_csum'],16)
        self.assertEqual(diff_csum_data_time_point_series[3].data['another_value_diff_csum'],10)


    def test_derivative_integral(self):
        
        # Test data        
        data_time_slot_series = DataTimeSlotSeries()
        data_time_slot_series.append(DataTimeSlot(start=TimePoint(0), end=TimePoint(1), data={'value':10}))
        data_time_slot_series.append(DataTimeSlot(start=TimePoint(1), end=TimePoint(2), data={'value':12}))
        data_time_slot_series.append(DataTimeSlot(start=TimePoint(2), end=TimePoint(3), data={'value':15}))
        data_time_slot_series.append(DataTimeSlot(start=TimePoint(3), end=TimePoint(4), data={'value':16}))
        
        # Test standard derivativeative behavior, from the series
        derivative_data_time_slot_series = derivative(data_time_slot_series)
        self.assertEqual(len(derivative_data_time_slot_series), 4)
        self.assertAlmostEqual(derivative_data_time_slot_series[0].data['value_derivative'],2)     # 0
        self.assertAlmostEqual(derivative_data_time_slot_series[1].data['value_derivative'],2.5)   # 2
        self.assertAlmostEqual(derivative_data_time_slot_series[2].data['value_derivative'],2)     # 5
        self.assertAlmostEqual(derivative_data_time_slot_series[3].data['value_derivative'],1.0)   # 6

        # Test integral as well  
        derivative_integral_data_time_slot_series = integral(derivative_data_time_slot_series, c=10)
        self.assertEqual(derivative_integral_data_time_slot_series[0].data['value_derivative_integral'], 10)
        self.assertEqual(derivative_integral_data_time_slot_series[1].data['value_derivative_integral'], 12)
        self.assertEqual(derivative_integral_data_time_slot_series[2].data['value_derivative_integral'], 15)
        self.assertEqual(derivative_integral_data_time_slot_series[3].data['value_derivative_integral'], 16)
        
        # Test data        
        data_time_slot_series = DataTimeSlotSeries()
        data_time_slot_series.append(DataTimeSlot(start=TimePoint(0), end=TimePoint(10), data={'value':10}))
        data_time_slot_series.append(DataTimeSlot(start=TimePoint(10), end=TimePoint(20), data={'value':12}))
        data_time_slot_series.append(DataTimeSlot(start=TimePoint(20), end=TimePoint(30), data={'value':15}))
        data_time_slot_series.append(DataTimeSlot(start=TimePoint(30), end=TimePoint(40), data={'value':16}))
        
        # Test standard derivativeative behavior, from the series
        derivative_data_time_slot_series = derivative(data_time_slot_series)
        self.assertEqual(len(derivative_data_time_slot_series), 4)
        self.assertAlmostEqual(derivative_data_time_slot_series[0].data['value_derivative'],0.2)
        self.assertAlmostEqual(derivative_data_time_slot_series[1].data['value_derivative'],0.25)
        self.assertAlmostEqual(derivative_data_time_slot_series[2].data['value_derivative'],0.2)
        self.assertAlmostEqual(derivative_data_time_slot_series[3].data['value_derivative'],0.1)        

        # Multi-key Test data
        data_time_point_series = DataTimePointSeries()
        data_time_point_series.append(DataTimePoint(t=0, data={'value':10, 'another_value': 75}))
        data_time_point_series.append(DataTimePoint(t=10, data={'value':12, 'another_value': 65}))
        data_time_point_series.append(DataTimePoint(t=20, data={'value':6, 'another_value': 32}))
        data_time_point_series.append(DataTimePoint(t=30, data={'value':16, 'another_value': 10}))
        derivative_data_time_point_series = data_time_point_series.derivative()
        self.assertAlmostEqual(derivative_data_time_point_series[0].data['value_derivative'],0.2)
        self.assertAlmostEqual(derivative_data_time_point_series[0].data['another_value_derivative'],-1.0)
        self.assertAlmostEqual(derivative_data_time_point_series[1].data['value_derivative'],-0.2)
        self.assertAlmostEqual(derivative_data_time_point_series[1].data['another_value_derivative'],-2.15)

        # Test derivative-integral identity with ther data     
        data_time_slot_series = DataTimeSlotSeries()
        data_time_slot_series.append(DataTimeSlot(start=TimePoint(0), end=TimePoint(1), data={'value':0}))
        data_time_slot_series.append(DataTimeSlot(start=TimePoint(1), end=TimePoint(2), data={'value':6}))
        data_time_slot_series.append(DataTimeSlot(start=TimePoint(2), end=TimePoint(3), data={'value':14}))
        data_time_slot_series.append(DataTimeSlot(start=TimePoint(3), end=TimePoint(4), data={'value':16}))
        data_time_slot_series.append(DataTimeSlot(start=TimePoint(4), end=TimePoint(5), data={'value':20}))
        
        derivative_data_time_slot_series = data_time_slot_series.derivative()
        derivative_integral_data_time_slot_series = derivative_data_time_slot_series.integral()

        self.assertAlmostEqual(derivative_integral_data_time_slot_series[0].data['value_derivative_integral'],0)
        self.assertAlmostEqual(derivative_integral_data_time_slot_series[1].data['value_derivative_integral'],6)
        self.assertAlmostEqual(derivative_integral_data_time_slot_series[2].data['value_derivative_integral'],14)
        self.assertAlmostEqual(derivative_integral_data_time_slot_series[3].data['value_derivative_integral'],16)
        self.assertAlmostEqual(derivative_integral_data_time_slot_series[4].data['value_derivative_integral'],20)


    def test_mavg(self):

        data_time_slot_series = DataTimeSlotSeries()
        data_time_slot_series.append(DataTimeSlot(start=TimePoint(0), end=TimePoint(1), data={'value':2}))
        data_time_slot_series.append(DataTimeSlot(start=TimePoint(1), end=TimePoint(2), data={'value':4}))
        data_time_slot_series.append(DataTimeSlot(start=TimePoint(2), end=TimePoint(3), data={'value':8}))
        data_time_slot_series.append(DataTimeSlot(start=TimePoint(3), end=TimePoint(4), data={'value':16}))
  
        mavg_data_time_slot_series = mavg(data_time_slot_series, 2)
        self.assertEqual(mavg_data_time_slot_series[0].data['value_mavg_2'], 3)
        self.assertEqual(mavg_data_time_slot_series[1].data['value_mavg_2'], 6)
        self.assertEqual(mavg_data_time_slot_series[2].data['value_mavg_2'], 12)
        self.assertEqual(mavg_data_time_slot_series[0].t, 1)
        self.assertEqual(mavg_data_time_slot_series[1].t, 2)
        self.assertEqual(mavg_data_time_slot_series[2].t, 3)


    def test_min_max_avg(self):
        
        # Test Python built-in default behaviour
        self.assertEqual(str(min), 'Min operation')
        self.assertEqual(min([1,2,3]),1)
        self.assertEqual(str(max), 'Max operation')
        self.assertEqual(max([1,2,3]),3)
        
        # Test data
        data_time_slot_series = DataTimeSlotSeries()
        data_time_slot_series.append(DataTimeSlot(start=TimePoint(0), end=TimePoint(60), data={'value':10}))
        data_time_slot_series.append(DataTimeSlot(start=TimePoint(60), end=TimePoint(120), data={'value':12}))
        data_time_slot_series.append(DataTimeSlot(start=TimePoint(120), end=TimePoint(180), data={'value':6}))
        data_time_slot_series.append(DataTimeSlot(start=TimePoint(180), end=TimePoint(240), data={'value':16}))
        
        # Test standalone
        self.assertEqual(min(data_time_slot_series), 6)
        self.assertEqual(max(data_time_slot_series), 16)
        self.assertEqual(avg(data_time_slot_series), 11)
        
        # Test from the series
        self.assertEqual(data_time_slot_series.min(), 6)
        self.assertEqual(data_time_slot_series.max(), 16)
        self.assertEqual(data_time_slot_series.avg(), 11)

        # Multi-key Test data
        data_time_point_series = DataTimePointSeries()
        data_time_point_series.append(DataTimePoint(t=0, data={'value':10, 'another_value': 75}))
        data_time_point_series.append(DataTimePoint(t=60, data={'value':12, 'another_value': 65}))
        data_time_point_series.append(DataTimePoint(t=120, data={'value':6, 'another_value': 32}))
        data_time_point_series.append(DataTimePoint(t=180, data={'value':16, 'another_value': 10}))
        
        self.assertEqual(data_time_point_series.min(), {'value': 6, 'another_value': 10})
        self.assertEqual(data_time_point_series.min(key='value'), 6)
        self.assertEqual(data_time_point_series.min(key='another_value'), 10)

        self.assertEqual(data_time_point_series.max(), {'value': 16, 'another_value': 75})
        self.assertEqual(data_time_point_series.max(key='value'), 16)        
        self.assertEqual(data_time_point_series.max(key='another_value'), 75)

        self.assertEqual(data_time_point_series.avg(), {'value': 11, 'another_value': 45.5})
        self.assertEqual(data_time_point_series.avg(key='value'), 11)          
        self.assertEqual(data_time_point_series.avg(key='another_value'), 45.5)


class TestSeriesOperations(unittest.TestCase):
    
    def test_filter(self):

        test_time_point_series = TimePointSeries(TimePoint(t=60),TimePoint(t=120),TimePoint(t=130))

        # Test from/to filtering
        self.assertEqual(len(filter(test_time_point_series, from_t=1, to_t=140)),3)

        # Test from/to filtering from the series
        self.assertEqual(len(test_time_point_series.filter(from_t=1, to_t=140)),3)
        self.assertEqual(len(test_time_point_series.filter(from_t=61, to_t=140)),2)
        self.assertEqual(len(test_time_point_series.filter(from_t=61)),2)
        self.assertEqual(len(test_time_point_series.filter(to_t=61)),1)

        # Test  data
        data_time_point_series =  DataTimePointSeries(DataTimePoint(t=60, data={'a':1, 'b':2}),
                                                      DataTimePoint(t=120, data={'a':2, 'b':4}),
                                                      DataTimePoint(t=180, data={'a':3, 'b':8}))

        # Test get item by string key (filtering on data labels)
        self.assertEqual(filter(data_time_point_series, 'b')[0].data, {'b': 2})

        # Test get item by string key (filtering on data labels), from the series
        self.assertEqual(data_time_point_series.filter('a')[0].data, {'a': 1})
        self.assertEqual(data_time_point_series.filter('a')[1].data, {'a': 2})
        self.assertEqual(data_time_point_series.filter('a')[2].data, {'a': 3})
        self.assertEqual(len(data_time_point_series.filter('a')), 3 )
        
        # Test that we haven't modified the original series
        self.assertEqual(data_time_point_series.data_keys(), ['a', 'b'])
        self.assertEqual(data_time_point_series[1].data['b'], 4)


    def test_select(self):
    
        # Test data
        data_time_point_series = DataTimePointSeries()
        data_time_point_series.append(DataTimePoint(t=0, data={'value':10, 'another_value': 75}))
        data_time_point_series.append(DataTimePoint(t=60, data={'value':12, 'another_value': 65}))
        data_time_point_series.append(DataTimePoint(t=120, data={'value':6, 'another_value': 32}))
        data_time_point_series.append(DataTimePoint(t=180, data={'value':16, 'another_value': 10}))
        
        # Basic select test        
        self.assertEqual(select(data_time_point_series, query='another_value=65')[0].t, 60)
        self.assertEqual(data_time_point_series.select('"another_value"=65')[0].t, 60)






# -*- coding: utf-8 -*-
"""Provides Units and TimeUnits (which fully support calendar arithmetic including DST changes)."""

import re
import datetime
import math
from .time import s_from_dt , dt_from_s, get_tz_offset_s
from .time import check_dt_consistency, correct_dt_dst
from .utilities import is_numerical
from .exceptions import ConsistencyException
            
# Setup logging
import logging
logger = logging.getLogger(__name__)


HARD_DEBUG = False


class Unit(object):
    """A generic unit."""
    
    def __init__(self, value):
        # TODO: check numerical or list? An perhaps support string representations?
        self.value = value
        
    def __repr__(self):   
        return str(self.value)

    def __add__(self, other):

        # Sum with numerical
        if is_numerical(self.value) and is_numerical(other):
            return self.value + other

        # Sum with another Unit
        if self.__class__ == other.__class__:
            return self.value + other.value

        # Sum with an objects with coordinates (Points, Slots)
        try:
            if len(other.coordinates) > 1:
                raise NotImplementedError('Cannot sum multidimensional objects')             
            return other.__class__(self.value + other.coordinates[0])
        except:
            raise NotImplementedError('Don\'t know how to add Units with {}'.format(other.__class__.__name__)) 
    
    def __radd__(self, other):
        return self.__add__(other)

    def __sub__(self, other):

        # Subtract with numerical
        if is_numerical(self.value) and is_numerical(other):
            return self.value - other

        # Subtract with another Unit
        if self.__class__ == other.__class__:
            return self.value - other.value

        # Sum with an objects with coordinates (Points, Slots)
        try:
            if len(other.coordinates) > 1:
                raise NotImplementedError('Cannot subtract multidimensional objects')             
            return other.__class__(self.value - other.coordinates[0])
        except:
            raise NotImplementedError('Don\'t know how to subtract Units with {}'.format(other.__class__.__name__)) 

    def __rsub__(self, other):
        return -self.__sub__(other)


    def __truediv__(self, other):

        # Divide with numerical
        if is_numerical(self.value) and is_numerical(other):
            return self.value / other
        else:
            if is_numerical(self.value):
                raise NotImplementedError('Division with "{}" is not implemented'.format(other.__class__.__name__))
            else:
                raise NotImplementedError('Cannot divide multidimensional objects')

    def __rtruediv__(self, other):
        return ( 1 / self.__truediv__(other))


    def __mul__(self, other):

        # Multiply with numerical
        if is_numerical(self.value) and is_numerical(other):
            return self.value * other
        else:
            if is_numerical(self.value):
                raise NotImplementedError('Division with "{}" is not implemented'.format(other.__class__.__name__))
            else:
                raise NotImplementedError('Cannot divide multidimensional objects')

    def __rmul__(self, other):
        return self.__mul__(other)
    
    def __eq__(self,other):
        if self is other:
            return True
        elif is_numerical(other):
            return self.value == other
        else:
            try:
                return self.value == other.value
            except AttributeError:
                return False


class TimeUnit(Unit):
    """A unit which can represent both physical (fixed) and calendar (variable) time units.
    Handles precision up to the microsecond and can be summed and subtracted with numerical
    values, Python datetime objects, other TimeUnits, or TimePoints.
    
    Can be initialized both using a numerical or string value, and by explicitly setting years,
    months, weeks, days, hours, minutes, seconds and microseconds. In the string representation,
    the mapping is as follows::
    
        {
          'Y': 'years',
          'M': 'months',
          'W': 'weeks',
          'D': 'days',
          'h': 'hours',
          'm': 'minutes',
          's': 'seconds',
          'u': 'microseconds'
        }
     
    So for example to create a time unit of one hour you can do a ``TimeUnit('1h')``, or to set
    it explicitly ``TimeUnit(hours=1)``. Also ``TimeUnit(3600)`` would work.
    """
    
    _CALENDAR = 'Calendar'
    _PHYSICAL = 'Physical'
    
    # NOT ref to https://docs.python.org/3/library/datetime.html#strftime-and-strptime-format-codes :  %d, %m, %w %y - %H, %M, %S
    # Instead: M, D, Y, W - h m s
    
    _mapping_table = { 
                       'Y': 'years',
                       'M': 'months',
                       'W': 'weeks',
                       'D': 'days',
                       'h': 'hours',
                       'm': 'minutes',
                       's': 'seconds',
                       'u': 'microseconds'
                      }

    def __init__(self, value=None, years=0, weeks=0, months=0, days=0, hours=0, minutes=0, seconds=0, microseconds=0, start=None, end=None, trustme=False):

        if not trustme:

            if value:
                if is_numerical(value):
                    string = '{}s'.format(value)
                else:
                    if not isinstance(value, str):
                        raise TypeError('TimeUnits must be initialized with a number, a string or explicitly setting years, months, days, hours etc. (Got "{}")'.format(string.__class__.__name__))
                    string = value
            else:
                string = None

            # Value OR explicit OR start/end
            if value and (years or months or days or hours or minutes or seconds or microseconds) and ((start is not None) or (end is not None)):
                raise ValueError('Choose between string/numerical init and explicit setting of years, months, days, hours etc.')
    
            # Check types:
            if not isinstance(years, int): raise ValueError('year not of type int (got "{}")'.format(years.__class__.__name__))
            if not isinstance(weeks, int): raise ValueError('weeks not of type int (got "{}")'.format(weeks.__class__.__name__))
            if not isinstance(months, int): raise ValueError('months not of type int (got "{}")'.format(months.__class__.__name__))
            if not isinstance(days, int): raise ValueError('days not of type int (got "{}")'.format(days.__class__.__name__))
            if not isinstance(hours, int): raise ValueError('hours not of type int (got "{}")'.format(hours.__class__.__name__))
            if not isinstance(minutes, int): raise ValueError('minutes not of type int (got "{}")'.format(minutes.__class__.__name__))
            if not isinstance(seconds, int): raise ValueError('seconds not of type int (got "{}")'.format(seconds.__class__.__name__))
            if not isinstance(microseconds, int): raise ValueError('microseconds not of type int (got "{}")'.format(microseconds.__class__.__name__))

            # Check that both start and end are set if one is set
            if start and not end:
                raise ValueError('You provided the start but not the end')
            if end and not start:
                raise ValueError('You provided the end but not the start')
                   
        # Set the TimeUnit in seconds
        if start and end:
            seconds = s_from_dt((end-start).dt)

        self.years        = years
        self.months       = months
        self.weeks        = weeks
        self.days         = days
        self.hours        = hours
        self.minutes      = minutes 
        self.seconds      = seconds
        self.microseconds = microseconds

        if string:

            # Specific case for floating point seconds (TODO: improve me, maybe inlcude in the regex?)            
            if string.endswith('s') and '.' in string:
                if '_' in string:
                    raise NotImplementedError('Composite TimeUnits with floating point seconds not yet implemented.')
                self.seconds = int(string.split('.')[0])
                
                # Get decimal seconds as string 
                decimal_seconds_str = string.split('.')[1][0:-1] # Remove the last "s"
                
                # Ensure we can handle precision
                if len(decimal_seconds_str) > 6:
                    decimal_seconds_str = decimal_seconds_str[0:6]
                    #raise ValueError('Sorry, "{}" has too many decimal seconds to be handled with a TimeUnit (which supports up to the microsecond).'.format(string))
                
                # Add missing trailing zeros
                missing_trailing_zeros = 6-len(decimal_seconds_str)                
                for _ in range(missing_trailing_zeros):
                    decimal_seconds_str += '0'
                
                # Cast to int & set
                self.microseconds = int(decimal_seconds_str)   

                return
            
            # Parse string using regex
            self.strings = string.split("_")
            regex = re.compile('^([0-9]+)([YMDWhmsu]{1,2})$')
            
            for string in self.strings:
                try:
                    groups   =  regex.match(string).groups()
                except AttributeError:
                    raise ValueError('Cannot parse string representation for the TimeUnit, unknown format ("{}")'.format(string)) from None

                setattr(self, self._mapping_table[groups[1]], int(groups[0]))

        if not trustme:  
                     
            # If nothing set, raise error
            if not self.years and not self.weeks and not self.months and not self.days and not self.hours and not self.minutes and not self.seconds and not self.microseconds:
                raise ValueError('Detected zero-duration TimeUnit!')
 
    def __repr__(self):
        string = ''
        if self.years: string += str(self.years)               + 'Y' + '_'
        if self.months: string += str(self.months)             + 'M' + '_'
        if self.weeks: string += str(self.weeks)               + 'W' + '_'
        if self.days: string += str(self.days)                 + 'D' + '_'
        if self.hours: string += str(self.hours)               + 'h' + '_'
        if self.minutes: string += str(self.minutes)           + 'm' + '_'
        if self.seconds: string += str(self.seconds)           + 's' + '_'
        if self.microseconds: string += str(self.microseconds) + 'u' + '_'

        string = string[:-1]
        return string

    def __add__(self, other):
        
        # TODO: This has to stay here or a circular import will start. Maybe refactor this?
        from .datastructures import TimePoint
        
        if isinstance(other, self.__class__):  
            return TimeUnit(years        = self.years + other.years,
                            months       = self.months + other.months,
                            weeks        = self.weeks + other.weeks,
                            days         = self.days + other.days,
                            hours        = self.hours + other.hours,
                            minutes      = self.minutes + other.minutes,
                            seconds      = self.seconds + other.seconds,
                            microseconds = self.microseconds + other.microseconds)

        elif isinstance(other, datetime.datetime):
            if not other.tzinfo:
                raise ValueError('Timezone of the datetime to sum with is required')
            return self.shift_dt(other, times=1)
        
        elif isinstance(other, TimePoint):
            return TimePoint(dt=self.shift_dt(other.dt, times=1))
        
        elif is_numerical(other):
            return other + self.value
            
        else:
            raise NotImplementedError('Adding TimeUnits with objects of class "{}" is not implemented'.format(other.__class__.__name__))
   
    def __radd__(self, other):
        return self.__add__(other)

    def __rsub__(self, other):
        
        # TODO: This has to stay here or a circular import will start. Maybe refactor this?
        from .datastructures import TimePoint
        
        if isinstance(other, self.__class__):
            raise NotImplementedError('Subracting a TimeUnit from another TimeUnit is not implemented to prevent negative TimeUnits.')


        elif isinstance(other, datetime.datetime):
            if not other.tzinfo:
                raise ValueError('Timezone of the datetime to sum with is required')
            return self.shift_dt(other, times=-1)
        
        
        elif isinstance(other, TimePoint):
            return TimePoint(dt=self.shift_dt(other.dt, times=-1))

        elif is_numerical(other):
            return other - self.value

        else:
            raise NotImplementedError('Subracting TimeUnits with objects of class "{}" is not implemented'.format(other.__class__.__name__))
   
    def __sub__(self, other):
        raise NotImplementedError('Cannot subrtact anything from a TimeUnit. Only a TimeUnit from something else.')

    def __eq__(self, other):

        # Check for the quickest first (value)
        try:
            if self.value == other:
                return True
        except TypeError:
            # Raised if this TimeUnit is of calendar type
            pass
        
        # Check against another Unit with value
        if isinstance(other, Unit):
            try:
                if self.value == other.value:
                    return True
            except TypeError:
                # Raised if this or the other other TimeUnit is of calendar type
                pass
          
        # Check against another TimeUnit using datetime components
        if isinstance(other, TimeUnit):
            if self.years != other.years:
                return False
            if self.months != other.months:
                return False
            if self.weeks != other.weeks:
                return False
            if self.days != other.days:
                return False
            if self.hours != other.hours:
                return False
            if self.minutes != other.minutes:
                return False
            if self.seconds != other.seconds:
                return False
            if self.microseconds != other.microseconds:
                return False
            return True
        
        # If everything fails, return false:
        return False

    def _is_composite(self):
        types = 0
        for item in self._mapping_table:
            if getattr(self, self._mapping_table[item]): types +=1
        return True if types > 1 else False 

    @property
    def value(self):
        """The value of the TimeUnit, in seconds. Not defined for calendar time units (as they have variable duration)."""
        try:
            return self._value
        except AttributeError:
            if self.type == self._CALENDAR:
                raise TypeError('Sorry, the value of a calendar TimeUnit is not defined. use duration_s() providing the starting point.') from None
            self._value = self.duration_s()
            return self._value

    @property
    def type(self):
        """The type of the TimeUnit.
        
           - "Physical" if based on hours, minutes, seconds, micorseconds (fixed duration)
           - "Calendar" if based on years, months, weeks, days (variable duration depending on the starting date,
             math not always well defined (30 Jan + 1 month does not make sense)."""

        if self.years or self.months or self.weeks or self.days:
            return self._CALENDAR
        elif self.hours or self.minutes or self.seconds or self.microseconds:
            return self._PHYSICAL
        else:
            raise ConsistencyException('Error, TimeSlot not initialized?!')
    
    def is_physical(self):
        """Return True if the TimeUnit type is physical, False otherwise"""
        if self.type == self._PHYSICAL:
            return True
        else:
            return False
        
    def is_calendar(self):
        """Return True if the TimeUnit type is calendar, False otherwise"""
        if self.type == self._CALENDAR:
            return True 
        else:       
            return False        

    def round_dt(self, time_dt, how = None):
        """Round a datetime according to this TimeUnit."""

        if self._is_composite():
            raise ValueError('Sorry, only simple TimeUnits are supported by the rebase operation')

        if not time_dt.tzinfo:
            raise ValueError('Timezone of the datetime is required')    
                
        # Handle physical time 
        if self.type == self._PHYSICAL:
        
            # Convert input time to seconds
            time_s = s_from_dt(time_dt)
            tz_offset_s = get_tz_offset_s(time_dt)
            
            # Get TimeUnit duration in seconds
            time_unit_s = self.duration_s(time_dt)

            # Apply modular math (including timezone time translation trick if required (multiple hours))
            # TODO: check for correctness, the time shift should be always done...
            
            if self.hours > 1 or self.minutes > 60:
                time_floor_s = ( (time_s - tz_offset_s) - ( (time_s - tz_offset_s) % time_unit_s) ) + tz_offset_s
            else:
                time_floor_s = time_s - (time_s % time_unit_s)
                
            time_ceil_s   = time_floor_s + time_unit_s
            
            if how == 'floor':
                time_rounded_s = time_floor_s
             
            elif how == 'ceil':
                time_rounded_s = time_ceil_s
            
            else:

                distance_from_time_floor_s = abs(time_s - time_floor_s) # Distance from floor
                distance_from_time_ceil_s  = abs(time_s - time_ceil_s)  # Distance from ceil

                if distance_from_time_floor_s < distance_from_time_ceil_s:
                    time_rounded_s = time_floor_s
                else:
                    time_rounded_s = time_ceil_s
                
            rounded_dt = dt_from_s(time_rounded_s, tz=time_dt.tzinfo)

        # Handle calendar time 
        elif self.type == self._CALENDAR:
            
            if self.years:
                if self.years > 1:
                    raise NotImplementedError('Cannot round based on calendar TimeUnits with years > 1')
                rounded_dt=time_dt.replace(month=1, day=1, hour=0, minute=0, second=0, microsecond=0)
                
            if self.months:
                if self.months > 1:
                    raise NotImplementedError('Cannot round based on calendar TimeUnits with months > 1')
                rounded_dt=time_dt.replace(day=1, hour=0, minute=0, second=0, microsecond=0)

            if self.weeks:
                raise NotImplementedError('Cannot round based on calendar TimeUnits with weeks')

            if self.days:
                if self.days > 1:
                    raise NotImplementedError('Cannot round based on calendar TimeUnits with days > 1')
                rounded_dt=time_dt.replace(hour=0, minute=0, second=0, microsecond=0)

            # Check DST offset consistency and fix if not respected
            if not check_dt_consistency(rounded_dt):
                rounded_dt = correct_dt_dst(rounded_dt)
            
            if how == 'ceil':
                rounded_dt = self.shift_dt(rounded_dt, 1)

        # Handle other cases (Consistency error)
        else:
            raise ConsistencyException('Error, TimeSlot type not physical nor calendar?!')

        # Return
        return rounded_dt

    def floor_dt(self, time_dt):
        """Floor a datetime according to this TimeUnit."""       
        return self.round_dt(time_dt, how='floor')
     
    def ceil_dt(self, time_dt):
        """Ceil a datetime according to this TimeUnit."""        
        return self.round_dt(time_dt, how='ceil')

    def rebase_dt(self, time_dt):
        """Rebase a given datetime to this TimeUnit."""
        return self.round_dt(time_dt, how='floor')
              
    def shift_dt(self, time_dt, times=1):
        """Shift a given datetime of n times of this TimeUnit."""
        if self._is_composite():
            raise ValueError('Sorry, only simple TimeUnits are supported by the rebase operation')
 
        # Convert input time to seconds
        time_s = s_from_dt(time_dt)
         
        # Handle physical time TimeSlot
        if self.type == self._PHYSICAL:
            
            # Get TimeUnit duration in seconds
            time_unit_s = self.duration_s()

            time_shifted_s = time_s + ( time_unit_s * times )
            time_shifted_dt = dt_from_s(time_shifted_s, tz=time_dt.tzinfo)
            
            return time_shifted_dt   

        # Handle calendar time TimeSlot
        elif self.type == self._CALENDAR:
            
            if times != 1:
                raise NotImplementedError('Cannot shift calendar TimeUnits for times than 1 (got times="{}")'.format(times))
            
            # Create a TimeDelta object for everything but years and months
            delta = datetime.timedelta(weeks = self.weeks,
                                       days = self.days,
                                       hours = self.hours,
                                       minutes = self.minutes,
                                       seconds = self.seconds,
                                       microseconds = self.microseconds)
            
            # Apply the time deta for the shif
            time_shifted_dt = time_dt + delta
            
            # Handle years
            if self.years:
                time_shifted_dt = time_shifted_dt.replace(year=time_shifted_dt.year + self.years)
            
            # Handle months
            if self.months:
                
                tot_months = self.months + time_shifted_dt.month             
                years_to_add = math.floor(tot_months/12.0)
                new_month = (tot_months % 12 )
                if new_month == 0:
                    new_month=12
                    years_to_add = years_to_add -1
    
                time_shifted_dt = time_shifted_dt.replace(year=time_shifted_dt.year + years_to_add)
                time_shifted_dt = time_shifted_dt.replace(month=new_month)
            
            # Check DST offset consistency and fix if not respected
            if not check_dt_consistency(time_shifted_dt):
                time_shifted_dt = correct_dt_dst(time_shifted_dt)

        # Handle other cases (Consistency error)
        else:
            raise ConsistencyException('Error, TimeSlot type not physical nor calendar?!')
        
        # Return
        return time_shifted_dt   

    def duration_s(self, start_dt=None):
        """Get the duration of the TimeUnit in seconds."""

        if self.type == self._CALENDAR:

            if not start_dt:
                raise ValueError('With a calendar TimeUnit you can ask for duration only if you provide the starting point')

            if self._is_composite():
                raise ValueError('Sorry, only simple TimeUnits are supported by this operation')

            # Start Epoch
            start_epoch = s_from_dt(start_dt)

            # End epoch
            end_dt = self.shift_dt(start_dt, 1)
            end_epoch = s_from_dt(end_dt)

            # Get duration based on seconds
            time_unit_s = end_epoch - start_epoch

        elif self.type == 'Physical':
            time_unit_s = 0
            if self.hours:
                time_unit_s += self.hours * 60 * 60
            if self.minutes:
                time_unit_s += self.minutes * 60
            if self.seconds:
                time_unit_s += self.seconds
            if self.microseconds:
                time_unit_s += 1/1000000.0 * self.microseconds
        
        else:
            raise ConsistencyException('Unknown TimeUnit type "{}"'.format(self.type))
        
        return time_unit_s

    # Get start/end/center
    def _get_start(self, end=None, center=None):
        new_values = []
        if end is not None:
            for i in range(len(self.value)):
                new_values.append(end.values[i] - self.value[i])
            return end.__class__(labels=end.labels, values=new_values, tz=end.tz)        
        elif center is not None:
            for i in range(len(self.value)):
                new_values.append(center.values[i] - self.value[i]/2.0)
            return center.__class__(labels=center.labels, values=new_values, tz=center.tz)
        else:
            raise ValueError('get_start: Got not end nor center')        
            
    def _get_end(self, start=None, center=None):
        new_values = []
        if start is not None:
            for i in range(len(self.value)):
                new_values.append(self.value[i] + start.values[i])
            return start.__class__(labels=start.labels, values=new_values, tz=start.tz)           
        elif center is not None:
            for i in range(len(self.value)):
                new_values.append(center.values[i] + self.value[i]/2.0) 
            return center.__class__(labels=center.labels, values=new_values, tz=center.tz) 
        else:
            raise ValueError('get_end: Got not end nor center')

    def _get_center(self, start=None, end=None):
        new_values = []
        if start is not None:
            for i in range(len(self.value)):
                new_values.append(start.values[i] + self.value[i]/2.0)
            return start.__class__(labels=start.labels, values=new_values, tz=start.tz)           
        elif end is not None:
            for i in range(len(self.value)):
                new_values.append(end.values[i] - self.value[i]/2.0) 
            return end.__class__(labels=end.labels, values=new_values, tz=end.tz) 
        else:
            raise ValueError('get_center: Got not end not start') 



# -*- coding: utf-8 -*-
"""Plotting utilities."""
# Note: the code in this module is spaghetti-ish code, and there are no tests. A major refactoring is required.'''

import os
import uuid
import datetime
from .time import dt_from_s, dt_to_str, dt_from_str
from .datastructures import DataTimePointSeries, DataTimeSlotSeries
from .units import TimeUnit
from .utilities import is_numerical

# Setup logging
import logging
logger = logging.getLogger(__name__)

AGGREGATE_THRESHOLD = 10000
render_mark_as_index = False

#=================
#   Utilities
#=================

def _utc_fake_s_from_dt(dt):
    dt_str = dt_to_str(dt)
    if '+' in dt_str:
        return s_from_dt(dt_from_str(dt_str.split('+')[0]+'+00:00'))
    elif '-' in dt_str:
        return s_from_dt(dt_from_str(dt_str.split('-')[0]+'+00:00'))
    else:
        raise ValueError('Cannot convert to fake UTC epoch a datetime without a timezone')



def _get_keys_and_check_data_for_plot(data):
    if is_numerical(data):
        keys = []
    elif isinstance(data, list):
        keys = []
        for key,item in enumerate(data): 
            if not is_numerical(item):
                raise Exception('Don\'t know how to plot item "{}" of type "{}"'.format(item, item.__class__.__name__))
            keys.append(key)
    
    elif isinstance(data, dict):
        keys = []
        for key,item in data.items(): 
            if not is_numerical(item):
                raise Exception('Don\'t know how to plot item "{}" of type "{}"'.format(item, item.__class__.__name__))
            keys.append(key)            
    else:
        raise Exception('Don\'t know how to plot data "{}" of type "{}"'.format(data, data.__class__.__name__))
    return keys

from timeseria.time import s_from_dt
def _to_dg_time(dt):
    '''Get Dygraphs time form datetime'''
    #return '{}{:02d}-{:02d}T{:02d}:{:02d}:{:02d}+00:00'.format(dt.year, dt.month, dt.day, dt.hour, dt.minute,dt.second)
    #return s_from_dt(dt)*1000
    if dt.microsecond:
        return 'new Date(Date.UTC({}, {}, {}, {}, {}, {}, {}))'.format(dt.year, dt.month-1, dt.day, dt.hour, dt.minute, dt.second, dt.microsecond/1000)
    else:
        return 'new Date(Date.UTC({}, {}, {}, {}, {}, {}))'.format(dt.year, dt.month-1, dt.day, dt.hour, dt.minute, dt.second)


def _to_dg_data(serie,  indexes_to_plot, aggregate_by=0):
    '''{% for timestamp,value in metric_data.items %}{{timestamp}},{{value}}\n{%endfor%}}'''
    
    dg_data=''
    first_t  = None
    last_t   = None

    global_min = None
    global_max = None
    
    if serie.mark:
        serie_mark_start_t = s_from_dt(serie.mark[0])
        serie_mark_end_t = s_from_dt(serie.mark[1])

    for i, item in enumerate(serie):
        
        # Offset
        i=i+1
                
        # Prepare a dict for every piece of data
        try:
            keys
        except UnboundLocalError:
            # Set keys and support vars
            keys = _get_keys_and_check_data_for_plot(item.data)
            if not keys:
                keys = [None]
            data_sums = [0 for key in keys]
            data_mins = [None for key in keys]
            data_maxs = [None for key in keys]
            index_sums = [0 for index in indexes_to_plot]
            

        #====================
        #  Aggregation
        #====================        
        if aggregate_by:

            # Set first timestamp
            if first_t is None:
                first_t = item.t
                  
            # Define data
            if keys:
                datas=[]      
                for j, key in enumerate(keys):
                    if key is None:
                        datas.append(item.data)
                    else:
                        datas.append(item.data[key])
            else:
                datas = [item.data]
            
            
            # Loop over data keys, including the "None" key if we have no keys (float data), and add data
            for j, data in enumerate(datas):
                
                # Sum
                data_sums[j] += data
                
                # Global min
                if global_min is None:
                    global_min = data
                else:
                    if data < global_min:
                        global_min = data                
                
                # Global max
                if global_max is None:
                    global_max = data
                else:
                    if data > global_max:
                        global_max = data

                # Min
                if data_mins[j] is None:
                    data_mins[j] = data
                else:
                    if data < data_mins[j]:
                        data_mins[j] = data
    
                # Max
                if data_maxs[j] is None:
                    data_maxs[j] = data
                else:
                    if data > data_maxs[j]:
                        data_maxs[j] = data

            # Loop over series indexes and add data
            for j, index in enumerate(indexes_to_plot):
                # TODO: plot only some indexes, now we plot all of them
                try:
                    index_value = getattr(item, index)
                    if index_value is not None:
                        index_sums[j] += index_value
                except AttributeError:
                    pass

            # Dump aggregated data?
            if (i!=0)  and ((i % aggregate_by)==0):
                if isinstance(serie, DataTimePointSeries):
                    last_t = item.t
                elif isinstance(serie, DataTimeSlotSeries):
                    last_t = item.end.t
                aggregation_t = first_t + ((last_t-first_t) /2)
                data_part=''
                
                # Data
                for i, key in enumerate(keys):
                    avg = data_sums[i]/aggregate_by
                    #if i  in index_metrics_positions:
                    #    data_part+='[{},{},{}],'.format( 0, avg, avg)
                    #else:
                    data_part+='[{},{},{}],'.format( data_mins[i], avg, data_maxs[i])

                # Indexes
                for i, index in enumerate(indexes_to_plot):
                    if index_sums[i] is not None:
                        data_part+='[0,{0},{0}],'.format(index_sums[i]/aggregate_by)
                    else:
                        data_part+='[,,],'

                # Do we have a mark?
                if serie.mark and render_mark_as_index:
                    if item.start.t >= serie_mark_start_t and item.end.t < serie_mark_end_t:                                    
                        # Add the (active) mark
                        data_part+='[0,1,1],'
                    else:
                        # Add the (inactive) mark
                        data_part+='[0,0,0],'
                
                # Remove last comma
                data_part = data_part[0:-1]
                
                # Add to dg_data
                dg_data += '[{},{}],'.format(_to_dg_time(dt_from_s(aggregation_t, tz=item.tz)), data_part)
                
                # Reset averages
                data_sums = [0 for key in keys]
                data_mins = [None for key in keys]
                data_maxs = [None for key in keys]
                index_sums = [0 for index in indexes_to_plot]
                first_t = None

        #====================
        #  No aggregation
        #====================
        else:
                    
            # Loop over data keys, including the "None" key if we have no keys (float data), and add data
            data_part=''
            for key in keys:
                if key is None:
                    data = item.data
                else:
                    data = item.data[key]

                data_part += '{},'.format(data)   #item.data
                
                # Global min
                if global_min is None:
                    global_min = data
                else:
                    if data < global_min:
                        global_min = data                
                
                # Global max
                if global_max is None:
                    global_max = data
                else:
                    if data > global_max:
                        global_max = data
                
            # Loop over series indexes and add data
            for index in indexes_to_plot:
                # TODO: plot only some indexes, now we plot all of them
                try:
                    index_value = getattr(item, index)
                    if index_value is not None:
                        data_part+='{},'.format(index_value)
                    else:
                        data_part+=','
                except AttributeError:
                    data_part+=','

            # Do we have a mark?
            if serie.mark and render_mark_as_index:
                if item.t >= serie_mark_start_t and item.t < serie_mark_end_t:
                    # Add the (active) mark
                    data_part+='1,'
                else:
                    # Add the (inactive) mark
                    data_part+='0,'

            # Remove last comma
            data_part = data_part[0:-1]

            # Add to dg_data
            dg_data += '[{},{}],'.format(_to_dg_time(dt_from_s(item.t, tz=item.tz)), data_part)

    if dg_data.endswith(','):
        dg_data = dg_data[0:-1]

    return global_min, global_max, '['+dg_data+']'


#=================
#  Dygraphs plot
#=================

def dygraphs_plot(timeseries, indexes=None, aggregate=None, aggregate_by=None, log_js=False):
    """Plot a timeseries in Jupyter using Dygraphs. Based on the work here: https://www.stefaanlippens.net/jupyter-custom-d3-visualization.html"""
    from IPython.display import display, Javascript, HTML
    
    if len(timeseries)==0:
        raise Exception('Cannot plot empty timeseries')
   
    if aggregate_by is None:
        # Set default aggregate_by if not explicitly set to something
        if len(timeseries)  > AGGREGATE_THRESHOLD:
            aggregate_by = 10**len(str(int(len(timeseries)/float(AGGREGATE_THRESHOLD))))
        else:
            aggregate_by = None
 
    if aggregate is not None:
        # Handle the case where we are required not to aggregate

        if not aggregate:
            aggregate_by = None

    #if show_data_reconstructed:
    #    show_data_loss=False

    # Checks
    if isinstance(timeseries, DataTimePointSeries):
        stepPlot_value   = 'false'
        drawPoints_value = 'true'
        if aggregate_by:
            legend_pre = 'Point (aggregated by {}) at '.format(aggregate_by)
        else:
            legend_pre = 'Point at '
    elif isinstance(timeseries, DataTimeSlotSeries):
        if isinstance(timeseries._resolution, TimeUnit):
            serie_unit_string = str(timeseries._resolution)
        else:
            serie_unit_string = str(timeseries._resolution) #.split('.')[0]+'s'
        #if aggregate_by: 
        #    raise NotImplementedError('Plotting slots with a plot-level aggregation is not yet supported')
        stepPlot_value   = 'true'
        drawPoints_value = 'false'
        if aggregate_by:
            # TODO: "Slot of {} unit" ?
            legend_pre = 'Slot of {}x{} (aggregated) starting at '.format(aggregate_by, serie_unit_string)
        else:
            legend_pre = 'Slot of {} starting at '.format(serie_unit_string)

    else:
        raise Exception('Don\'t know how to plot an object of type "{}"'.format(timeseries.__class__.__name__))

    # Set title
    if timeseries.title:
        title = timeseries.title
    else:
        if isinstance(timeseries, DataTimePointSeries):
            if aggregate_by:
                title = 'Time series of #{} points at {}, aggregated by {}'.format(len(timeseries), timeseries._resolution_string, aggregate_by)
            else:
                title = 'Time series of #{} points at {}'.format(len(timeseries), timeseries._resolution_string)
             
        elif isinstance(timeseries, DataTimeSlotSeries):
            if aggregate_by:
                # TODO: "slots of unit" ?
                title = 'Time series of #{} slots of {}, aggregated by {}'.format(len(timeseries), serie_unit_string, aggregate_by)
            else:
                # TODO: "slots of unit" ?
                title = 'Time series of #{} slots of {}'.format(len(timeseries), serie_unit_string)
                
        else:
            title = timeseries.__class__.__name__

    if aggregate_by:
        logger.info('Aggregating by "{}" for improved plotting'.format(aggregate_by))
            
    # Define div name
    graph_id = 'graph_' + str(uuid.uuid4()).replace('-', '')
    graph_div_id  = graph_id + '_plotarea'
    legend_div_id = graph_id + '_legend'

    # Do we have to show a series mark?
    if timeseries.mark:
        logger.info('Found series mark and showing it')
        serie_mark_html = '<span style="background:rgba(255, 255, 102, .6);">&nbsp;mark&nbsp;</span>'
        serie_mark_html_off = '<span style="background:rgba(255, 255, 102, .2);">&nbsp;mark&nbsp;</span>'
    else:
        serie_mark_html=''
        serie_mark_html_off = ''

    # Dygraphs javascript
    dygraphs_javascript = """
// Taken and adapted from dygraph.js
function legendFormatter(data) {
      var g = data.dygraph;

      if (g.getOption('showLabelsOnHighlight') !== true) return '';

      var sepLines = g.getOption('labelsSeparateLines');
      var html;

      if (typeof data.x === 'undefined') {
        if (g.getOption('legend') != 'always') {
          return '';
        }
        // We are here when there's no selection and {legend: 'always'} is set.
        html = '"""+title+""": ';
        for (var i = 0; i < data.series.length; i++) {
          var series = data.series[i];
          // Skip not visible series
          if (!series.isVisible) continue;
          
          // Also skip Timeseria-specific series
          if (series.label=='data_mark') continue;
          
          // Add some initial stuff
          if (html !== '') html += sepLines ? '<br/>' : ' ';
          

          if ((series.label=='data_reconstructed') || (series.label=='data_loss') || (series.label=='forecast') || (series.label=='anomaly')){
              html += "<span style='margin-left:15px; background: " + series.color + ";'>&nbsp;" + series.labelHTML + "&nbsp</span>, ";
          }
          else {
              html += "<span style='margin-left:15px; font-weight: bold; color: " + series.color + ";'>" + series.dashHTML + " " + series.labelHTML + "</span>, ";
          }
          
          // Remove last comma and space
          html = html.substring(0, html.length - 2);
        }
        html += ' &nbsp;&nbsp; """+serie_mark_html +"""'
        return html;
      }

      // Find out if we have an active mark
      mark_active = false
      for (var i = 0; i < data.series.length; i++) {
        if ( data.series[i].label=='data_mark' && data.series[i].y==1 )  {
           mark_active = true;
        }
      }


      html = '""" + legend_pre + """' + data.xHTML + ' (""" + str(timeseries.tz)+ """):';
      for (var i = 0; i < data.series.length; i++) {
        
        var mark_active = false;
        
        if ( data.series[i].label=='data_mark' && data.series[i].y==1 )  {
           mark_active = true;
        }
                 
        var series = data.series[i];
        
        // Skip not visible series
        if (!series.isVisible) continue;
        
        // Also skip Timeseria-specific series
        if (series.label=='data_mark') continue;
        
        if (sepLines) html += '<br>';
        //var decoration = series.isHighlighted ? ' class="highlight"' : '';
        var decoration = series.isHighlighted ? ' style="background-color: #000000"' : '';
        
        /*
        if (series.isHighlighted) {
            decoration = ' style="background-color: #fcf8b0"'
          }
        else{
            decoration = ' style="background-color: #ffffff"'
        }
        console.log(decoration)
        */
        
        //decoration = ' style="background-color: #fcf8b0"'
        if ((series.label=='data_reconstructed') || (series.label=='data_loss') || (series.label=='forecast') || (series.label=='anomaly')){
            html += "<span" + decoration + "> <span style='background: " + series.color + ";'>&nbsp" + series.labelHTML + "&nbsp</span>:&#160;" + series.yHTML*100 + "%</span>, ";
        
        }
        else {
            html += "<span" + decoration + "> <b><span style='color: " + series.color + ";'>" + series.labelHTML + "</span></b>:&#160;" + series.yHTML + "</span>, ";
        }
      
      }
      // Remove last comma and space
      html = html.substring(0, html.length - 2);
      
      if (mark_active) {
          html += ' &nbsp;"""+ serie_mark_html +"""'
      }
      //else {
      //    html += ' &nbsp;"""+ serie_mark_html_off +"""'
      //}
      return html;
    };
"""
    dygraphs_javascript += 'new Dygraph('
    dygraphs_javascript += 'document.getElementById("{}"),'.format(graph_div_id)

    # Loop over all the keys
    labels=''
    keys = _get_keys_and_check_data_for_plot(timeseries[0].data)
    if keys:
        for key in keys:
            if isinstance(key, int):
                if len(keys) == 1:
                    labels='value,'
                else:    
                    labels += 'value {},'.format(key+1)
            else:
                labels += '{},'.format(key)
        # Remove last comma
        labels = labels[0:-1]
    else:
        labels='value'

    # Handle series indexes
    if indexes is None:
        # Plot alls eries indexes
        indexes_to_plot = timeseries.indexes
    else:
        # Check that the indexes are of the right type and that are present in the series indexes
        indexes_to_plot = []
        if not isinstance(indexes, list):
            raise TypeError('The "indexes" argument must be a list')
        for index in indexes:
            if not isinstance(index, str):
                raise TypeError('The "indexes" list items must be string (got type "{}")'.format(index.__class__.__name__))
            if not index in timeseries.indexes:
                raise ValueError('The index "{}" is not present in the series indexes ({})'.format(index, timeseries.indexes))
            indexes_to_plot.append(index)
    
    # Set index labels
    for index in indexes_to_plot:
        labels+=',{}'.format(index)

    # Handle series mark (as index)
    if timeseries.mark and render_mark_as_index:
        labels+=',data_mark'
    
    # Prepare labels string list for Dygraphs
    labels_list = ['Timestamp'] + labels.split(',')

    # Get series data in Dygraphs format (and aggregate if too much data and find global min and max)
    global_min, global_max, dg_data = _to_dg_data(timeseries, indexes_to_plot, aggregate_by)
    if global_max != global_min:
        plot_min = str(global_min-((global_max-global_min)*0.1))
        plot_max = str(global_max+((global_max-global_min)*0.1))
    else:
        # Case where there is a straight line
        plot_min = str(global_min*0.99)
        plot_max = str(global_max*1.01)

    #dygraphs_javascript += '"Timestamp,{}\\n{}",'.format(labels, dg_data)
    dygraphs_javascript += '{},'.format(dg_data)

#dateWindow: [0,"""+str(_utc_fake_s_from_dt(timeseries[-1].end.dt)*1000)+"""],

    dygraphs_javascript += """{\
labels: """+str(labels_list)+""",
drawGrid: true,
drawPoints:"""+drawPoints_value+""",
strokeWidth: 1.5,
pointSize:1.5,
highlightCircleSize:4,
stepPlot: """+stepPlot_value+""",
fillGraph: false,
fillAlpha: 0.5,
colorValue: 0.5,
showRangeSelector: true,
//rangeSelectorHeight: 30,
hideOverlayOnMouseOut: true,
interactionModel: Dygraph.defaultInteractionModel,
includeZero: false,
showRoller: false,
legend: 'always',
labelsUTC: true,
legendFormatter: legendFormatter,
labelsDiv: '"""+legend_div_id+"""',
valueRange: ["""+plot_min+""", """+plot_max+"""],
zoomCallback: function() {
    this.updateOptions({zoomRange: ["""+plot_min+""", """+plot_max+"""]});
},
axes: {
  /*y: {
    // set axis-related properties here
    drawGrid: false,
    independentTicks: true
  },*/
  y2: {
    drawGrid: false,
    drawAxis: false,
    axisLabelWidth:0,
    independentTicks: true,
    valueRange: [0,1.01],
    //customBars: false, // Does not work?
  }/*,
  x : {
                    valueFormatter: Dygraph.dateString_,
                    valueParser: function(x) { return new Date(x); },
 }*/
},
animatedZooms: true,"""

    # Define colors in case of an aggregated plot and not. Has to be
    # differentiated due different transparency policies in Dygraphs
    if aggregate_by:        
        rgba_value_red    = 'rgba(255,128,128,0.4)'   # Alpha is for the legend
        rgba_value_gray   = 'rgba(240,240,240,0.5)'   # Alpha is for the legend
        rgba_value_orange = 'rgba(235, 156, 56,1.0)'  # Alpha is for the legend
        fill_alpha_value  = 0.31                      # Alpha for the area 
    else:
        rgba_value_red    = 'rgba(255,128,128,0.4)'   # Alpha is for the legend
        rgba_value_gray   = 'rgba(240,240,240,0.5)'   # Alpha is for the legend
        rgba_value_orange = 'rgba(245, 193, 130,0.8)' # Alpha is for the legend
        fill_alpha_value  = 0.5                       # Alpha for the area

    # Fixed fill alpha value
    fill_alpha_value_fixed = 0.6

    # Fixed colors
    rgba_alpha_violet = 'rgba(227, 168, 237, 0.5)'   #  235, 179, 245 | 227, 156, 240
    rgba_value_yellow = 'rgba(255, 255, 102, 0.6)'
    rgba_value_darkorange = 'rgba(255, 140, 0, 0.7)'
    
    # Special series start
    dygraphs_javascript += """
     series: {"""
    
    # Data reconstructed index series
    if 'data_reconstructed' in indexes_to_plot:
        dygraphs_javascript += """
       'data_reconstructed': {
         //customBars: false, // Does not work?
         axis: 'y2',
         drawPoints: false,
         strokeWidth: 0,
         highlightCircleSize:0,
         fillGraph: true,
         fillAlpha: """+str(fill_alpha_value_fixed)+""",   // This alpha is used for the area
         color: '"""+rgba_alpha_violet+"""'                // Alpha here is used for the legend 
       },"""
    
    # Data loss index series
    if 'data_loss' in indexes_to_plot:
        # Add data loss special timeseries
        dygraphs_javascript += """
       'data_loss': {
         //customBars: false, // Does not work?
         axis: 'y2',
         //stepPlot: true,
         drawPoints: false,
         strokeWidth: 0,
         highlightCircleSize:0,
         fillGraph: true,
         fillAlpha: """+str(fill_alpha_value)+""",  // This alpha is used for the area 
         color: '"""+rgba_value_red+"""'            // Alpha here is used for the legend 
       },"""

    # Data forecast index series
    if 'forecast' in indexes_to_plot:
        dygraphs_javascript += """
       'forecast': {
         //customBars: false, // Does not work?
         axis: 'y2',
         //stepPlot: true,
         drawPoints: false,
         strokeWidth: 0,
         highlightCircleSize:0,
         fillGraph: true,
         fillAlpha: """+str(fill_alpha_value)+""",  // This alpha is used for the area 
         color: '"""+rgba_value_yellow+"""'         // Alpha here is used for the legend 
       },"""

    # Add anomaly index series
    if 'anomaly' in indexes_to_plot:
        dygraphs_javascript += """
        'anomaly': {
         //customBars: false, // Does not work?
         axis: 'y2',
         drawPoints: false,
         strokeWidth: 0,
         highlightCircleSize:0,
         fillGraph: true,
         fillAlpha: """+str(fill_alpha_value_fixed)+""",  // This alpha is used for the area
         color: '"""+rgba_value_darkorange+"""'           // Alpha here is used for the legend 
       },"""
   
    # Add data mark index series
    dygraphs_javascript += """
       'data_mark': {
         //customBars: false, // Does not work?
         axis: 'y2',
         drawPoints: false,
         strokeWidth: 0,
         highlightCircleSize:0,
         fillGraph: true,
         fillAlpha: """+str(fill_alpha_value)+""",  // This alpha is used for the area
         color: '"""+rgba_value_yellow+"""'         // Alpha here is used for the legend 
       },"""

    # Add all non-index series to be included in the miniplot
    for label in timeseries.data_keys():
        dygraphs_javascript += """
           '"""+str(label)+"""': {
             showInRangeSelector:true
           },"""     
    
    # Special series end
    dygraphs_javascript += """
     },"""

    # Force "original" Dygraph color if only one data series           
    if len(timeseries.data_keys()) <=1:
        dygraphs_javascript += """colors: ['rgb(0,128,128)'],""" 

    # Handle series mark (only if not handled as an index)
    if timeseries.mark and not render_mark_as_index:
 
        # Check that we know how to use this mark
        if not isinstance(timeseries.mark[0], datetime.datetime):
            raise TypeError('Series marks must be datetime objects')
        if not isinstance(timeseries.mark[1], datetime.datetime):
            raise TypeError('Series marks must be datetime objects')            
           
        # Convert the mark to fake epoch milliseconds
        mark_start = _utc_fake_s_from_dt(timeseries.mark[0])*1000
        mark_end   = _utc_fake_s_from_dt(timeseries.mark[1])*1000
         
        # Add js code
        dygraphs_javascript  += 'underlayCallback: function(canvas, area, g) {'
        dygraphs_javascript += 'var bottom_left = g.toDomCoords({}, -20);'.format(mark_start)
        dygraphs_javascript += 'var top_right = g.toDomCoords({}, +20);'.format(mark_end)
        dygraphs_javascript += 'var left = bottom_left[0];'
        dygraphs_javascript += 'var right = top_right[0];'
        dygraphs_javascript += 'canvas.fillStyle = "rgba(255, 255, 102, .5)";'
        dygraphs_javascript += 'canvas.fillRect(left, area.y, right - left, area.h);'
        dygraphs_javascript += '}'

    if aggregate_by:
        dygraphs_javascript += 'customBars: true'        
    dygraphs_javascript += '})'
    
    if log_js:
        logger.info(dygraphs_javascript)

    rendering_javascript = """
require.undef('"""+graph_id+"""'); // Helps to reload in Jupyter
define('"""+graph_id+"""', ['dgenv'], function (Dygraph) {

    function draw(div_id) {
    """ + dygraphs_javascript +"""
    }
    return draw;
});
"""

    STATIC_DATA_PATH = '/'.join(os.path.realpath(__file__).split('/')[0:-1]) + '/static/'

    # Require Dygraphs Javascript library
    # https://raw.githubusercontent.com/sarusso/Timeseria/develop/timeseria/static/js/dygraph-2.1.0.min
    display(Javascript("require.config({paths: {dgenv: 'https://cdnjs.cloudflare.com/ajax/libs/dygraph/2.1.0/dygraph.min'}});"))
    # TODO: load it locally, maybe with something like:    
    #with open(STATIC_DATA_PATH+'js/dygraph-2.1.0.min.js') as dy_js_file:
    #    display(Javascript(dy_js_file.read()))

    # Load Dygraphs CSS
    with open(STATIC_DATA_PATH+'css/dygraph-2.1.0.css') as dg_css_file:
        display(HTML('<style>'+dg_css_file.read()+'</style>'))
    
    # Load Dygraphs Javascript and html code    
    display(Javascript(rendering_javascript))
    display(HTML('''<div style="height:36px; padding:0; margin-left:0px; margin-top:10px">
                    <div id="'''+legend_div_id+'''" style="width:100%"></div></div>
                    <div id="'''+graph_div_id+'''" style="width:100%; margin-right:0px"></div>'''))

    # Render the graph (this is the "entrypoint")
    display(Javascript("""
        (function(element){
            require(['"""+graph_id+"""'], function("""+graph_id+""") {
                """+graph_id+"""(element.get(0), '%s');
            });
        })(element);
    """ % (graph_div_id)))


#=================
# Matplotlib plot
#=================

def matplotlib_plot(timeseries):
    """Plot a timeseries in Jupyter using Matplotlib."""
    from matplotlib import pyplot
    pyplot.plot(timeseries.df)
    pyplot.show()



#!/usr/bin/env python

from setuptools import setup, find_packages

setup(name='timeseria',
      version='0.1.3',
      description='A time series processing library',
      url="https://github.com/sarusso/timeseria",
      author='Stefano Alberto Russo',
      author_email='stefano.russo@gmail.com',
      packages=['timeseria','timeseria.tests', 'timeseria.models'],
      package_data={
          'timeseria': ['static/css/*.css', 'static/js/*.js'],
          'timeseria.tests': ['test_data/csv/*.csv']
       },
      install_requires = [
                          'Keras >=2.1.3, <3.0.0',
                          'tensorflow >=1.15.2, <3.0.0',
                          'matplotlib >=2.1.2, <4.0.0',
                          'numpy >=1.19.5, <2.0.0',
                          'scikit-learn >=0.2.2, <2.0.0',
                          'pandas >=0.23.4, <2.0.0',
                          'chardet >=3.0.4, <4.0.0',
                          'convertdate >=2.1.2, <3.0.0',
                          'lunarcalendar >=0.0.9, <1.0.0',
                          'holidays >=0.10.3, <1.0.0',
                          'cython >=0.29.17, <1.0.0',
                          'requests >=2.20.0, <3.0.0',
                          'h5py >=2.10.0, <4.0.0',
                          'scipy >=1.5.4, <2.0.0'
                          ],
      license='Apache License 2.0',
      license_files = ('LICENSE',),
    )



import unittest
import os
from ..datastructures import DataTimePoint, DataTimePointSeries
from ..transformations import Resampler, Slotter 
from ..time import dt, s_from_dt, dt_from_str
from ..units import TimeUnit

# Setup logging
import logging
logging.basicConfig(level=os.environ.get('LOGLEVEL') if os.environ.get('LOGLEVEL') else 'CRITICAL')

# Set test data path
TEST_DATA_PATH = '/'.join(os.path.realpath(__file__).split('/')[0:-1]) + '/test_data/'

class TestSlotter(unittest.TestCase):

    def setUp(self):       
        
        # All the following time series have point with validity=1m
        
        # Time series from 16:58:00 to 17:32:00 (Europe/Rome)
        self.data_time_point_series_1 = DataTimePointSeries()
        start_t = 1436022000 - 120
        for i in range(35):
            data_time_point = DataTimePoint(t = start_t + (i*60),
                                            data = {'temperature': 154+i})
            self.data_time_point_series_1.append(data_time_point)
        
        # Time series from 17:00:00 to 17:30:00 (Europe/Rome)
        self.data_time_point_series_2 = DataTimePointSeries()
        start_t = 1436022000
        for i in range(34):
            data_time_point = DataTimePoint(t    = start_t + (i*60),
                                            data = {'temperature': 154+i})
            self.data_time_point_series_2.append(data_time_point)

        # Time series from 17:00:00 to 17:20:00 (Europe/Rome)
        self.data_time_point_series_3 = DataTimePointSeries()
        start_t = 1436022000 - 120
        for i in range(23):
            data_time_point = DataTimePoint(t    = start_t + (i*60),
                                            data = {'temperature': 154+i})
            self.data_time_point_series_3.append(data_time_point) 
        
        # Time series from 17:10:00 to 17:30:00 (Europe/Rome)
        self.data_time_point_series_4 = DataTimePointSeries()
        start_t = 1436022000 + 600
        for i in range(21):
            data_time_point = DataTimePoint(t    = start_t + (i*60),
                                            data = {'temperature': 154+i})
            self.data_time_point_series_4.append(data_time_point)

        # Time series from 16:58:00 to 17:32:00 (Europe/Rome)
        self.data_time_point_series_5 = DataTimePointSeries()
        start_t = 1436022000 - 120
        for i in range(35):
            if i > 10 and i <21:
                continue
            data_time_point = DataTimePoint(t    = start_t + (i*60),
                                            data = {'temperature': 154+i})
            self.data_time_point_series_5.append(data_time_point)

        # The following time series has point with validity=15m

        # Time series from 2019,10,1,1,0,0 to 2019,10,1,6,0,0 (Europe/Rome)
        from_dt  = dt(2019,10,1,1,0,0, tzinfo='Europe/Rome')
        to_dt    = dt(2019,10,1,6,0,0, tzinfo='Europe/Rome')
        time_unit = TimeUnit('15m') 
        self.data_time_point_series_6 = DataTimePointSeries()
        slider_dt = from_dt
        count = 0
        while slider_dt < to_dt:
            if count not in [1, 6, 7, 8, 9, 10]:
                data_time_point = DataTimePoint(t    = s_from_dt(slider_dt),
                                              data = {'temperature': 154+count},
                                              tz   = 'Europe/Rome')
                self.data_time_point_series_6.append(data_time_point)
            slider_dt = slider_dt + time_unit
            count += 1

        # The following time series has point with validity=1h

        # Time series from 2019,10,24,0,0,0 to 2019,10,31,0,0,0 (Europe/Rome), DST off -> 2 AM repeated
        from_dt   = dt(2019,10,24,0,0,0, tzinfo='Europe/Rome')
        to_dt     = dt(2019,10,31,0,0,0, tzinfo='Europe/Rome')
        time_unit = TimeUnit('1h') 
        self.data_time_point_series_7 = DataTimePointSeries()
        slider_dt = from_dt
        count = 0
        while slider_dt < to_dt:
            if count not in []:
                data_time_point = DataTimePoint(t    = s_from_dt(slider_dt),
                                                data = {'temperature': 154+count},
                                                tz   = 'Europe/Rome')
                self.data_time_point_series_7.append(data_time_point)
            slider_dt = slider_dt + time_unit
            count += 1


    def test_slot(self):
          
        # Time series from 16:58:00 to 17:32:00 (Europe/Rome), no from/to, extremes excluded (default)
        data_time_slot_series = Slotter('600s').process(self.data_time_point_series_1)        
        self.assertEqual(len(data_time_slot_series), 3)
        self.assertEqual(data_time_slot_series[0].start.dt, dt_from_str('2015-07-04 17:00:00+02:00'))
        self.assertEqual(data_time_slot_series[1].start.dt, dt_from_str('2015-07-04 17:10:00+02:00'))
        self.assertEqual(data_time_slot_series[2].start.dt, dt_from_str('2015-07-04 17:20:00+02:00'))
         
        # Time series from 16:58:00 to 17:32:00 (Europe/Rome), no from/to, extremes included.
        data_time_slot_series = Slotter('10m').process(self.data_time_point_series_1,  include_extremes = True)
        self.assertEqual(len(data_time_slot_series), 5)
        self.assertEqual(data_time_slot_series[0].start.dt, dt_from_str('2015-07-04 16:50:00+02:00'))
        self.assertEqual(data_time_slot_series[1].start.dt, dt_from_str('2015-07-04 17:00:00+02:00'))
        self.assertEqual(data_time_slot_series[2].start.dt, dt_from_str('2015-07-04 17:10:00+02:00'))
        self.assertEqual(data_time_slot_series[3].start.dt, dt_from_str('2015-07-04 17:20:00+02:00'))
        # The following is because the include_extremes is set to True
        self.assertEqual(data_time_slot_series[4].start.dt, dt_from_str('2015-07-04 17:30:00+02:00'))
        self.assertEqual(data_time_slot_series[4].data_loss, 0.75)
 
        # Time series from 16:58:00 to 17:32:00 (Europe/Rome), from 15:50 to 17:40
        data_time_slot_series = Slotter('10m').process(self.data_time_point_series_1,
                                    from_t = s_from_dt(dt_from_str('2015-07-04 16:50:00+02:00')),
                                    to_t   = s_from_dt(dt_from_str('2015-07-04 17:40:00+02:00')))
        self.assertEqual(len(data_time_slot_series), 5)
        self.assertEqual(data_time_slot_series[0].start.dt, dt_from_str('2015-07-04 16:50:00+02:00'))
        self.assertEqual(data_time_slot_series[1].start.dt, dt_from_str('2015-07-04 17:00:00+02:00'))
        self.assertEqual(data_time_slot_series[2].start.dt, dt_from_str('2015-07-04 17:10:00+02:00'))
        self.assertEqual(data_time_slot_series[3].start.dt, dt_from_str('2015-07-04 17:20:00+02:00'))
        self.assertEqual(data_time_slot_series[4].start.dt, dt_from_str('2015-07-04 17:30:00+02:00'))

        # Time series from 16:58:00 to 17:32:00 (Europe/Rome), from 15:00 to 18:00
        data_time_slot_series = Slotter('10m').process(self.data_time_point_series_1,
                                    from_t = s_from_dt(dt_from_str('2015-07-04 15:00:00+02:00')),
                                    to_t   = s_from_dt(dt_from_str('2015-07-04 18:00:00+02:00')))
        # Here from and to are capped with the time series data points
        self.assertEqual(str(data_time_slot_series[0].start.dt), str('2015-07-04 14:50:00+00:00'))
        self.assertEqual(str(data_time_slot_series[-1].start.dt), str('2015-07-04 15:30:00+00:00'))


        # Time series from 2019,10,1,1,0,0 to 2019,10,1,6,0,0 (Europe/Rome), validity=15m
        data_time_slot_series = Slotter('1h').process(self.data_time_point_series_6)
        self.assertEqual(len(data_time_slot_series), 4)
        self.assertEqual(str(data_time_slot_series[0].start.dt), str('2019-10-01 01:00:00+02:00'))
        self.assertEqual(str(data_time_slot_series[1].start.dt), str('2019-10-01 02:00:00+02:00'))
        self.assertEqual(str(data_time_slot_series[2].start.dt), str('2019-10-01 03:00:00+02:00'))
        self.assertEqual(str(data_time_slot_series[3].start.dt), str('2019-10-01 04:00:00+02:00'))
 
        # Test changing timezone of the series
        self.data_time_point_series_6.tz = 'America/New_York'
        data_time_slot_series = Slotter('1h').process(self.data_time_point_series_6)
        self.assertEqual(len(data_time_slot_series), 4)
        self.assertEqual(str(data_time_slot_series[0].start.dt), str('2019-09-30 19:00:00-04:00'))
        self.assertEqual(str(data_time_slot_series[1].start.dt), str('2019-09-30 20:00:00-04:00'))
        self.assertEqual(str(data_time_slot_series[2].start.dt), str('2019-09-30 21:00:00-04:00'))
        self.assertEqual(str(data_time_slot_series[3].start.dt), str('2019-09-30 22:00:00-04:00'))
         
        # Time series from 2019,10,24,0,0,0 to 2019,10,31,0,0,0 (Europe/Rome), DST off -> 2 AM repeated
        data_time_slot_series = Slotter('1h').process(self.data_time_point_series_7)
 
        self.assertEqual(len(data_time_slot_series), 168)
        self.assertEqual(str(data_time_slot_series[73].start.dt), str('2019-10-27 01:00:00+02:00'))
        self.assertEqual(str(data_time_slot_series[74].start.dt), str('2019-10-27 02:00:00+02:00'))
        self.assertEqual(str(data_time_slot_series[75].start.dt), str('2019-10-27 02:00:00+01:00'))
        self.assertEqual(str(data_time_slot_series[76].start.dt), str('2019-10-27 03:00:00+01:00'))
        self.assertEqual(data_time_slot_series[76].data['temperature_avg'],230.5)

        # Time series from 2019,10,24,0,0,0 to 2019,10,31,0,0,0 (Europe/Rome), DST off -> 2 AM repeated
        data_time_slot_series = Slotter('1D').process(self.data_time_point_series_7)
        self.assertEqual(len(data_time_slot_series), 6)
        self.assertEqual(str(data_time_slot_series[0].start.dt), str('2019-10-24 00:00:00+02:00'))
        self.assertEqual(str(data_time_slot_series[-1].start.dt), str('2019-10-29 00:00:00+01:00')) # Last one not included as right excluded


        # Downsampling (downslotting)
         
        # Use the time series 7 from 2019,10,24,0,0,0 to 2019,10,31,0,0,0 (Europe/Rome), DST off -> 2 AM repeated
        # Work in progress in the following
        #if False:
        #    print('--------------------------------------------')
        #    for item in self.data_time_point_series_7: print(item)
        #    print('--------------------------------------------')
        #    for i, item in enumerate(data_time_slot_series): print(i, item)
        #    print('--------------------------------------------')
        # This is a "downsampling", and there are data losses and strange values that should not be there.
        # TODO: fix me!
        #slotter = Slotter('10m')
        #data_time_slot_series = slotter.process(self.data_time_point_series_7, from_t=1571868600, to_t=1571873400)
        #for item in data_time_slot_series:
        #    print(item)
  
        # This is a 1-day slotting over a DST change, and slots have the wrong start time (but are  correct in legth).
        # TODO: fix me!
        #slotter = Slotter('1D')
        #data_time_slot_series = slotter.process(self.data_time_point_series_7)
        #for item in data_time_slot_series:
        #    print(item)
        
        # TODO: also test from/to with timezone and epoch.
        # i.e. 1571868000 will not be recognised as correct Europe/Rome midnight.
 
 
    def test_slot_operations(self):
        
        # Import (overloaded) operations
        from ..operations import min, max, avg, sum

        # Time series from 16:58:00 to 17:32:00 (Europe/Rome)
        data_time_point_series = DataTimePointSeries()
        start_t = 1436022000 - 120
        for i in range(35):
            data_time_point = DataTimePoint(t = start_t + (i*60),
                                            data = {'temperature': 154+i, 'humidity': 5})
            data_time_point_series.append(data_time_point)

        # Add extra operations
        data_time_slot_series = Slotter('600s', extra_operations=[min,max]).process(data_time_point_series)        
        self.assertEqual(data_time_slot_series[0].data, {'temperature_avg': 161.0, 'humidity_avg': 5.0, 'temperature_min': 156, 'humidity_min': 5, 'temperature_max': 166, 'humidity_max': 5})

        # Change default operations
        data_time_slot_series = Slotter('600s', default_operation=sum).process(data_time_point_series)        
        self.assertEqual(data_time_slot_series[0].data, {'temperature_sum': 1771, 'humidity_sum': 55})
        self.assertEqual(data_time_slot_series[1].data, {'temperature_sum': 1881, 'humidity_sum': 55})

        # Only extra operations
        data_time_slot_series = Slotter('600s', default_operation=None, extra_operations=[avg,min,max]).process(data_time_point_series)        
        self.assertEqual(data_time_slot_series[0].data, {'temperature_avg': 161.0, 'humidity_avg': 5.0, 'temperature_min': 156, 'humidity_min': 5, 'temperature_max': 166, 'humidity_max': 5})

        # Time series from 16:58:00 to 17:32:00 (Europe/Rome) with a short (single-slot) gap to be interpolated
        data_time_point_series = DataTimePointSeries()
        start_t = 1436022000 - 120
        for i in range(35):
            if i < 12 or i > 22:
                data_time_point = DataTimePoint(t = start_t + (i*60),
                                                data = {'temperature': 154+i, 'humidity': 5})
                data_time_point_series.append(data_time_point)

        data_time_slot_series = Slotter('600s', extra_operations=[min,max]).process(data_time_point_series)        
        self.assertEqual(data_time_slot_series[1].data['temperature_min'], 166.5)
        
        # Time series from 16:58:00 to 17:32:00 (Europe/Rome) with a long (multi-slot) gap to be interpolated
        data_time_point_series = DataTimePointSeries()
        start_t = 1436022000 - 120
        for i in range(60):
            if i < 10 or i >48:
                data_time_point = DataTimePoint(t = start_t + (i*60),
                                                data = {'temperature': 154+i, 'humidity': 5})
                data_time_point_series.append(data_time_point)

        data_time_slot_series = Slotter('600s', extra_operations=[min,max]).process(data_time_point_series)        
        self.assertEqual(data_time_slot_series[1].data['temperature_min'], 167.75)
        self.assertEqual(data_time_slot_series[2].data['temperature_min'], 179.5)
        self.assertEqual(data_time_slot_series[3].data['temperature_min'], 191.25)

        # Re-slotting with the same time unit as the original points
        data_time_slot_series = Slotter('60s').process(data_time_point_series)

        #  Upsampling (upslotting) not supported yet
        data_time_slot_series = Slotter('60s').process(data_time_point_series)
        with self.assertRaises(ValueError):    
            data_time_slot_series = Slotter('30s').process(data_time_point_series)        
        
        # TODO: directly load a day-resolution time series in this test
        data_time_slot_series = Resampler(86400).process(self.data_time_point_series_7)
        with self.assertRaises(ValueError):    
            Slotter('1h').process(data_time_slot_series)

    def test_slot_indexes(self):
        
        # Import (overloaded) operations
        from ..operations import min, max, avg, sum

        # Time series from 16:58:00 to 17:32:00 (Europe/Rome)
        data_time_point_series = DataTimePointSeries()
        start_t = 1436022000 - 120
        for i in range(35):
            if i > 10 and i < 15:
                anomaly = 0.25
                forecast = 0
                data_reconstructed = 1
                data_loss = 1
            elif i > 20:
                anomaly = None 
                forecast = 1
                data_reconstructed = 0
                data_loss = None
            else:
                anomaly = 0 
                forecast = 0
                data_reconstructed = 0
                data_loss = 0.1
                
            data_time_point = DataTimePoint(t = start_t + (i*60),
                                            data = {'temperature': 154+i, 'humidity': 5},
                                            data_loss = data_loss
                                            )
            # Add extra indexes as if something operated on the time series
            data_time_point.anomaly = anomaly
            data_time_point.forecast = forecast
            data_time_point._data_reconstructed = data_reconstructed
            
            # Append
            data_time_point_series.append(data_time_point)
                
        # This is an indirect test of the series indexes. TODO: move it away.
        self.assertEqual(data_time_point_series.indexes, ['data_reconstructed', 'data_loss', 'anomaly', 'forecast'])
        
        # Slot the time series
        slotted_data_time_point_series = Slotter('600s').process(data_time_point_series)

        # Check that we have all the indexes
        self.assertEqual(slotted_data_time_point_series.indexes, ['data_reconstructed', 'data_loss', 'anomaly', 'forecast'])

        # Check indexes math
        self.assertAlmostEqual(slotted_data_time_point_series[0].anomaly, 0.045454545454545456)
        self.assertAlmostEqual(slotted_data_time_point_series[0].data_reconstructed, 0.18181818181818182)
        self.assertAlmostEqual(slotted_data_time_point_series[0].forecast, 0)
        
        self.assertAlmostEqual(slotted_data_time_point_series[1].anomaly, 0.08333333333333333)
        self.assertAlmostEqual(slotted_data_time_point_series[1].data_reconstructed, 0.2727272727272727)
        self.assertAlmostEqual(slotted_data_time_point_series[1].forecast, 0.18181818181818182)

        self.assertEqual(slotted_data_time_point_series[2].anomaly, None)
        self.assertAlmostEqual(slotted_data_time_point_series[2].data_reconstructed, 0)
        self.assertEqual(slotted_data_time_point_series[2].forecast, 1.0)

        # The following is re-computed from missing coverage only, and not marked as None.
        # See the compute_data_loss function fore more details and some more comments.
        self.assertEqual(slotted_data_time_point_series[2].data_loss, 0) 


class TestResampler(unittest.TestCase):


    def test_resample(self):
        
        # Time series from 16:58:00 to 17:32:00 (Europe/Rome)
        data_time_point_series = DataTimePointSeries()
        start_t = 1436022000 - 120
        for i in range(35):
            data_time_point = DataTimePoint(t = start_t + (i*60),
                                            tz='Europe/Rome',
                                            data = {'temperature': 154+i})
            data_time_point_series.append(data_time_point)
 
        data_time_slot_series = Resampler('600s').process(data_time_point_series)        
 
        self.assertEqual(str(data_time_slot_series[0].dt), '2015-07-04 17:00:00+02:00')
        self.assertAlmostEqual(data_time_slot_series[0].data_loss, 0.25)
        self.assertEqual(data_time_slot_series[0].data['temperature'], 157.5)
         
        self.assertEqual(str(data_time_slot_series[1].dt), '2015-07-04 17:10:00+02:00')
        self.assertAlmostEqual(data_time_slot_series[1].data_loss, 0.0)
        self.assertEqual(data_time_slot_series[1].data['temperature'], 166.0)
         
        self.assertEqual(str(data_time_slot_series[3].dt), '2015-07-04 17:30:00+02:00')
        self.assertAlmostEqual(data_time_slot_series[3].data_loss, 0.25)
        self.assertEqual(data_time_slot_series[3].data['temperature'], 184.5)
 
     
        # Time series from 15:00:00 to 15:37:00 (UTC)
        data_time_point_series = DataTimePointSeries()
        start_t = 1436022000
        for i in range(38):
            data_time_point = DataTimePoint(t = start_t + (i*60),
                                            data = {'temperature': 154+i})
            data_time_point_series.append(data_time_point)
 
 
        data_time_slot_series = Resampler('600s').process(data_time_point_series)
         
        self.assertEqual(str(data_time_slot_series[0].dt), '2015-07-04 15:00:00+00:00')
        self.assertAlmostEqual(data_time_slot_series[0].data_loss, 0.45)
        self.assertEqual(data_time_slot_series[0].data['temperature'], 156.5)
         
        self.assertEqual(str(data_time_slot_series[1].dt), '2015-07-04 15:10:00+00:00')
        self.assertAlmostEqual(data_time_slot_series[1].data_loss, 0.0)
        self.assertEqual(data_time_slot_series[1].data['temperature'], 164.0)
         
        self.assertEqual(str(data_time_slot_series[4].dt), '2015-07-04 15:40:00+00:00')
        self.assertAlmostEqual(data_time_slot_series[4].data_loss, 0.85)
        self.assertEqual(data_time_slot_series[4].data['temperature'], 189.5)        
         
 
        # Time series from 00:00 to 00:19 (UTC) 
        data_time_point_series = DataTimePointSeries()
        start_t = 0
        for i in range(20):
            if i < 8 or i > 12:
                data_time_point = DataTimePoint(t = start_t + (i*60),
                                                data = {'temperature': 154+i})
                data_time_point_series.append(data_time_point)
 
        # Re-sample as is
        data_time_slot_series = Resampler('1m').process(data_time_point_series)        
        self.assertEqual(len(data_time_slot_series), 20)
         
        # Original points
        self.assertEqual(data_time_slot_series[0].data_loss, 0)
        self.assertEqual(data_time_slot_series[0].data['temperature'], 154)
        self.assertEqual(data_time_slot_series[1].data_loss, 0)
        self.assertEqual(data_time_slot_series[1].data['temperature'], 155)
        self.assertEqual(data_time_slot_series[19].data_loss, 0)
        self.assertEqual(data_time_slot_series[19].data['temperature'], 173)
 
        # Reconstructed (interpolated) points
        self.assertEqual(data_time_slot_series[8].data_loss, 1)
        self.assertEqual(data_time_slot_series[8].data['temperature'], 162)
        self.assertEqual(data_time_slot_series[9].data_loss, 1)
        self.assertEqual(data_time_slot_series[9].data['temperature'], 163)
        self.assertEqual(data_time_slot_series[12].data_loss, 1)
        self.assertEqual(data_time_slot_series[12].data['temperature'], 166)


        # Time series from 00:00 to 00:19 (UTC) with missing data in the middle
        data_time_point_series = DataTimePointSeries()
        start_t = 0
        for i in range(20):
            if i < 8 or i > 12:
                data_time_point = DataTimePoint(t = start_t + (i*60),
                                                data = {'temperature': 154+i})
                data_time_point_series.append(data_time_point)

        #  Upsampling not supported yet
        with self.assertRaises(ValueError):
            data_time_slot_series = Resampler('20s').process(data_time_point_series)        
        
        # When implementing it, remember about check correct linear interpolation when there is data loss vs. when there isn't,
        # and that all the prev-next math has to be correctly taken into account together with computing the right data loss.
         

    def test_resampler_indexes(self):
        
        # Import (overloaded) operations
        from ..operations import min, max, avg, sum

        # Time series from 16:58:00 to 17:32:00 (Europe/Rome)
        data_time_point_series = DataTimePointSeries()
        start_t = 1436022000 - 120
        for i in range(35):
            if i > 10 and i < 15:
                anomaly = 0.25
                forecast = 0
                data_reconstructed = 1
                data_loss = 1
            elif i > 20:
                anomaly = None 
                forecast = 1
                data_reconstructed = 0
                data_loss = None
            else:
                anomaly = 0 
                forecast = 0
                data_reconstructed = 0
                data_loss = 0.1
                
            data_time_point = DataTimePoint(t = start_t + (i*60),
                                            data = {'temperature': 154+i, 'humidity': 5},
                                            data_loss = data_loss
                                            )
            # Add extra indexes as if something operated on the time series
            data_time_point.anomaly = anomaly
            data_time_point.forecast = forecast
            data_time_point._data_reconstructed = data_reconstructed
            
            # Append
            data_time_point_series.append(data_time_point)
                
        # This is an indirect test of the series indexes. TODO: move it away.
        self.assertEqual(data_time_point_series.indexes, ['data_reconstructed', 'data_loss', 'anomaly', 'forecast'])
        
        # Slot the time series
        resampled_data_time_point_series = Resampler('600s').process(data_time_point_series)        

        # Check that we have all the indexes
        self.assertEqual(resampled_data_time_point_series.indexes, ['data_reconstructed', 'data_loss', 'anomaly', 'forecast'])

        # Check indexes math TODO: check the entire math here, there is something strange.
        self.assertAlmostEqual(resampled_data_time_point_series[0].anomaly, 0)
        self.assertAlmostEqual(resampled_data_time_point_series[0].data_reconstructed, 0)
        self.assertAlmostEqual(resampled_data_time_point_series[0].forecast, 0)
        
        self.assertAlmostEqual(resampled_data_time_point_series[1].anomaly, 0.09090909090909091)
        self.assertAlmostEqual(resampled_data_time_point_series[1].data_reconstructed, 0.36363636363636365)
        self.assertAlmostEqual(resampled_data_time_point_series[1].forecast, 0)

        self.assertEqual(resampled_data_time_point_series[2].anomaly, 0)
        self.assertAlmostEqual(resampled_data_time_point_series[2].data_reconstructed, 0)
        self.assertEqual(resampled_data_time_point_series[2].forecast, 0.6363636363636364)

        self.assertAlmostEqual(resampled_data_time_point_series[2].data_loss, 0.04) 

        # Now resample ath the same sampling interval and check indexes are still there
        same_resampled_data_time_point_series = data_time_point_series[0:5].resample(60)
        self.assertEqual(same_resampled_data_time_point_series.indexes, ['data_reconstructed', 'data_loss', 'anomaly', 'forecast'])




# -*- coding: utf-8 -*-
"""Data reconstructions models."""

import copy
import statistics
from ..utilities import get_periodicity, get_periodicity_index, set_from_t_and_to_t, item_is_in_range, mean_absolute_percentage_error
from ..time import dt_from_s
from sklearn.metrics import mean_absolute_error, mean_squared_error
from ..units import TimeUnit
from pandas import DataFrame
from math import sqrt

# Setup logging
import logging
logger = logging.getLogger(__name__)

# Suppress TensorFlow warnings as default behavior
import tensorflow as tf
tf.compat.v1.logging.set_verbosity(tf.compat.v1.logging.ERROR)

# Base models and utilities
from .base import TimeSeriesParametricModel, ProphetModel


#=========================
#  Generic Reconstructor
#=========================

class Reconstructor(TimeSeriesParametricModel):

    def _predict(self, *args, **kwargs):
        raise NotImplementedError('Reconstructors can be used only from the apply() method.') from None


    def _apply(self, timeseries, remove_data_loss=False, data_loss_threshold=1, inplace=False):
        logger.debug('Using data_loss_threshold="%s"', data_loss_threshold)

        # TODO: understand if we want the apply from/to behavior. For now it is disabled
        # (add from_t=None, to_t=None, from_dt=None, to_dt=None in the function call above)
        # from_t, to_t = set_from_t_and_to_t(from_dt, to_dt, from_t, to_t)
        # Maybe also add a timeseries.mark=[from_dt, to_dt]
         
        from_t = None
        to_t   = None
        
        if not inplace:
            timeseries = timeseries.duplicate()

        if len(timeseries.data_keys()) > 1:
            raise NotImplementedError('Multivariate time series are not yet supported')

        for key in timeseries.data_keys():
            
            gap_started = None
            
            for i, item in enumerate(timeseries):
                
                # Skip if before from_t/dt of after to_t/dt
                if from_t is not None and timeseries[i].t < from_t:
                    continue
                try:
                    # Handle slots
                    if to_t is not None and timeseries[i].end.t > to_t:
                        break
                except AttributeError:
                    # Handle points
                    if to_t is not None and timeseries[i].t > to_t:
                        break                

                if item.data_loss >= data_loss_threshold:
                    # This is the beginning of an area we want to reconstruct according to the data_loss_threshold
                    if gap_started is None:
                        gap_started = i
                else:
                    
                    if gap_started is not None:
                    
                        # Reconstruct for this gap
                        self._reconstruct(from_index=gap_started, to_index=i, timeseries=timeseries, key=key)
                        gap_started = None
                    
                    item._data_reconstructed = 0
                    
                if remove_data_loss:
                    item._data_loss = 0
            
            # Reconstruct the last gap as well if left "open"
            if gap_started is not None:
                self._reconstruct(from_index=gap_started, to_index=i+1, timeseries=timeseries, key=key)

        if not inplace:
            return timeseries
        else:
            return None


    def _evaluate(self, timeseries, steps='auto', limit=None, data_loss_threshold=1, metrics=['RMSE', 'MAE'], details=False, from_t=None, to_t=None, from_dt=None, to_dt=None):

        # Set evaluation_score steps if we have to
        if steps == 'auto':
            try:
                steps = [1, self.data['periodicity']]
            except KeyError:
                steps = [1, 2, 3]
        elif isinstance(steps, list):
            pass
        else:
            steps = list(range(1, steps+1))
         
        # Support vars
        evaluation_score = {}
        from_t, to_t = set_from_t_and_to_t(from_dt, to_dt, from_t, to_t)
        warned = False
        
        # Log
        logger.info('Will evaluate model for %s steps with metrics %s', steps, metrics)
        
        # Find areas where to evaluate the model
        for key in timeseries.data_keys():
             
            for steps_round in steps:
                
                # Support vars
                real_values = []
                reconstructed_values = []
                processed_samples = 0

                # Here we will have steps=1, steps=2 .. steps=n          
                logger.debug('Evaluating model for %s steps', steps_round)
                
                for i in range(len(timeseries)):

                    # Skip if needed
                    try:
                        if not item_is_in_range(timeseries[i], from_t, to_t):
                            continue
                    except StopIteration:
                        break                  
                
                    # Skip the first and the last ones, otherwise reconstruct the ones in the middle
                    if (i == 0) or (i >= len(timeseries)-steps_round):
                        continue

                    # Is this a "good area" where to test or do we have to stop?
                    stop = False
                    if timeseries[i-1].data_loss >= data_loss_threshold:
                        stop = True
                    for j in range(steps_round):
                        if timeseries[i+j].data_loss >= data_loss_threshold:
                            stop = True
                            break
                    if timeseries[i+steps_round].data_loss >= data_loss_threshold:
                        stop = True
                    if stop:
                        continue
                            
                    # Set prev and next
                    prev_value = timeseries[i-1].data[key]
                    next_value = timeseries[i+steps_round].data[key]
                    
                    # Compute average value
                    average_value = (prev_value+next_value)/2
                    
                    # Data to be reconstructed
                    timeseries_to_reconstruct = timeseries.__class__()
                    
                    # Append prev
                    #timeseries_to_reconstruct.append(copy.deepcopy(timeseries[i-1]))
                    
                    # Append in the middle and store real values
                    for j in range(steps_round):
                        item = copy.deepcopy(timeseries[i+j])
                        # Set the data_loss to one so the item will be reconstructed
                        item._data_loss = 1
                        item.data[key] = average_value
                        timeseries_to_reconstruct.append(item)
                        
                        real_values.append(timeseries[i+j].data[key])
              
                    # Append next
                    #timeseries_to_reconstruct.append(copy.deepcopy(timeseries[i+steps_round]))
                    
                    # Do we have a 1-point only timeseries? If so, manually set the resolution
                    # as otherwise it would be not defined. # TODO: does it make sense?
                    if len(timeseries_to_reconstruct) == 1:
                        timeseries_to_reconstruct._resolution = timeseries._resolution

                    # Apply model inplace
                    self._apply(timeseries_to_reconstruct, inplace=True)
                    processed_samples += 1

                    # Store reconstructed values
                    for j in range(steps_round):
                        reconstructed_values.append(timeseries_to_reconstruct[j].data[key])
                    
                    # Break if we have to
                    if limit is not None and processed_samples >= limit:
                        break
                    
                    # Warn if no limit given and we are over
                    if not limit and not warned and i > 10000:
                        logger.warning('No limit set in the evaluation with a quite long time series, this could take some time.')
                        warned=True
                        
                if limit and processed_samples < limit:
                    logger.warning('The evaluation limit is set to "{}" but I have only "{}" samples for "{}" steps'.format(limit, processed_samples, steps_round))

                if not reconstructed_values:
                    raise Exception('Could not evaluate model, maybe not enough data?')

                # Compute RMSE and ME, and add to the evaluation_score
                if 'RMSE' in metrics:
                    evaluation_score['RMSE_{}_steps'.format(steps_round)] = sqrt(mean_squared_error(real_values, reconstructed_values))
                if 'MAE' in metrics:
                    evaluation_score['MAE_{}_steps'.format(steps_round)] = mean_absolute_error(real_values, reconstructed_values)
                if 'MAPE' in metrics:
                    evaluation_score['MAPE_{}_steps'.format(steps_round)] = mean_absolute_percentage_error(real_values, reconstructed_values)

        # Compute overall RMSE
        if 'RMSE' in metrics:
            sum_rmse = 0
            count = 0
            for key in evaluation_score:
                if key.startswith('RMSE_'):
                    sum_rmse += evaluation_score[key]
                    count += 1
            evaluation_score['RMSE'] = sum_rmse/count

        # Compute overall MAE
        if 'MAE' in metrics:
            sum_me = 0
            count = 0
            for key in evaluation_score:
                if key.startswith('MAE_'):
                    sum_me += evaluation_score[key]
                    count += 1
            evaluation_score['MAE'] = sum_me/count

        # Compute overall MAPE
        if 'MAPE' in metrics:
            sum_me = 0
            count = 0
            for key in evaluation_score:
                if key.startswith('MAPE_'):
                    sum_me += evaluation_score[key]
                    count += 1
            evaluation_score['MAPE'] = sum_me/count
        
        if not details:
            simple_evaluation_score = {}
            if 'RMSE' in metrics:
                simple_evaluation_score['RMSE'] = evaluation_score['RMSE']
            if 'MAE' in metrics:
                simple_evaluation_score['MAE'] = evaluation_score['MAE']
            if 'MAPE' in metrics:
                simple_evaluation_score['MAPE'] = evaluation_score['MAPE']
            evaluation_score = simple_evaluation_score
            
        return evaluation_score


    def _reconstruct(self, *args, **krargs):
        raise NotImplementedError('Reconstruction for this model is not yet implemented')



#=========================
# P. Average Reconstructor
#=========================

class PeriodicAverageReconstructor(Reconstructor):
    
    def fit(self, timeseries, data_loss_threshold=0.5, periodicity=None, dst_affected=False, from_t=None, to_t=None, from_dt=None, to_dt=None, offset_method='average'):
        # This is a fit wrapper only to allow correct documentation
        """
        Fit the model. On top of the basic reconstructor parameters, also accepts:
 
        Args:
            timeseries (:obj:`timeseria.datastructures.DataTimePointSeries` or :obj:`timeseria.datastructures.DataTimeSlotSeries`): The timeseries on which to perform the reconstruction.
            data_loss_threshold (float): The threshold of the data loss to use to fit the model.
            periodicity (float, optional): used to force a specific periodicity.
         
        """
        return super(PeriodicAverageReconstructor, self).fit(timeseries, data_loss_threshold, periodicity, dst_affected, from_t, to_t, from_dt, to_dt, offset_method)

        
    
    def _fit(self, timeseries, data_loss_threshold=0.5, periodicity=None, dst_affected=False, from_t=None, to_t=None, from_dt=None, to_dt=None, offset_method='average'):

        if not offset_method in ['average', 'extremes']:
            raise Exception('Unknown offset method "{}"'.format(self.offset_method))
        self.offset_method = offset_method
    
        if len(timeseries.data_keys()) > 1:
            raise NotImplementedError('Multivariate time series are not yet supported')

        from_t, to_t = set_from_t_and_to_t(from_dt, to_dt, from_t, to_t)
        
        # Set or detect periodicity
        if periodicity is None:
            periodicity =  get_periodicity(timeseries)
            try:
                if isinstance(timeseries._resolution, TimeUnit):
                    logger.info('Detected periodicity: %sx %s', periodicity, timeseries._resolution)
                else:
                    logger.info('Detected periodicity: %sx %ss', periodicity, timeseries._resolution)
            except AttributeError:
                logger.info('Detected periodicity: %sx %ss', periodicity, timeseries._resolution)
                
        self.data['periodicity']  = periodicity
        self.data['dst_affected'] = dst_affected 
                
        for key in timeseries.data_keys():
            sums   = {}
            totals = {}
            processed = 0
            for item in timeseries:
                
                # Skip if needed
                try:
                    if not item_is_in_range(item, from_t, to_t):
                        continue
                except StopIteration:
                    break
                
                # Process
                if item.data_loss < data_loss_threshold:
                    periodicity_index = get_periodicity_index(item, timeseries._resolution, periodicity, dst_affected=dst_affected)
                    if not periodicity_index in sums:
                        sums[periodicity_index] = item.data[key]
                        totals[periodicity_index] = 1
                    else:
                        sums[periodicity_index] += item.data[key]
                        totals[periodicity_index] +=1
                processed += 1

        averages={}
        for periodicity_index in sums:
            averages[periodicity_index] = sums[periodicity_index]/totals[periodicity_index]
        self.data['averages'] = averages
        
        logger.debug('Processed "%s" items', processed)


    def _reconstruct(self, timeseries, key, from_index, to_index):
        logger.debug('Reconstructing between "{}" and "{}"'.format(from_index, to_index-1))

        # Compute offset (old approach)
        if self.offset_method == 'average':
            diffs=0
            for j in range(from_index, to_index):
                real_value = timeseries[j].data[key]
                periodicity_index = get_periodicity_index(timeseries[j], timeseries._resolution, self.data['periodicity'], dst_affected=self.data['dst_affected'])
                reconstructed_value = self.data['averages'][periodicity_index]
                diffs += (real_value - reconstructed_value)
            offset = diffs/(to_index-from_index)
        
        elif self.offset_method == 'extremes':
            # Compute offset (new approach)
            diffs=0
            try:
                for j in [from_index-1, to_index+1]:
                    real_value = timeseries[j].data[key]
                    periodicity_index = get_periodicity_index(timeseries[j], timeseries._resolution, self.data['periodicity'], dst_affected=self.data['dst_affected'])
                    reconstructed_value = self.data['averages'][periodicity_index]
                    diffs += (real_value - reconstructed_value)
                offset = diffs/2
            except IndexError:
                offset=0
        else:
            raise Exception('Unknown offset method "{}"'.format(self.offset_method))

        # Actually reconstruct
        for j in range(from_index, to_index):
            item_to_reconstruct = timeseries[j]
            periodicity_index = get_periodicity_index(item_to_reconstruct, timeseries._resolution, self.data['periodicity'], dst_affected=self.data['dst_affected'])
            item_to_reconstruct.data[key] = self.data['averages'][periodicity_index] + offset
            item_to_reconstruct._data_reconstructed = 1
                        

    def _plot_averages(self, timeseries, **kwargs):   
        averages_timeseries = copy.deepcopy(timeseries)
        for item in averages_timeseries:
            value = self.data['averages'][get_periodicity_index(item, averages_timeseries._resolution, self.data['periodicity'], dst_affected=self.data['dst_affected'])]
            if not value:
                value = 0
            item.data['average'] =value 
        averages_timeseries.plot(**kwargs)


#=========================
#  Prophet Reconstructor
#=========================

class ProphetReconstructor(Reconstructor, ProphetModel):
    
    def _fit(self, timeseries, from_t=None, to_t=None, from_dt=None, to_dt=None):

        from fbprophet import Prophet

        from_t, to_t = set_from_t_and_to_t(from_dt, to_dt, from_t, to_t)

        if len(timeseries.data_keys()) > 1:
            raise NotImplementedError('Multivariate time series are not yet supported')

        data = self.from_timeseria_to_prophet(timeseries, from_t, to_t)

        # Instantiate the Prophet model
        self.prophet_model = Prophet()
        
        # Fit tjhe Prophet model
        self.prophet_model.fit(data)


    def _reconstruct(self, timeseries, key, from_index, to_index):
        logger.debug('Reconstructing between "{}" and "{}"'.format(from_index, to_index-1))
    
        # Get and prepare data to reconstruct
        items_to_reconstruct = []
        for j in range(from_index, to_index):
            items_to_reconstruct.append(timeseries[j])
        data_to_reconstruct = [self.remove_timezone(dt_from_s(item.t)) for item in items_to_reconstruct]
        dataframe_to_reconstruct = DataFrame(data_to_reconstruct, columns = ['ds'])

        # Apply Prophet fit
        forecast = self.prophet_model.predict(dataframe_to_reconstruct)
        #forecast[['ds', 'yhat', 'yhat_lower', 'yhat_upper']].tail()

        # Ok, replace the values with the reconsturcted ones
        for i, j in enumerate(range(from_index, to_index)):
            #logger.debug('Reconstructing item #{} with reconstucted item #{}'.format(j,i))
            item_to_reconstruct = timeseries[j]
            item_to_reconstruct.data[key] = forecast['yhat'][i]
            item_to_reconstruct._data_reconstructed = 1

